"""Legacy interfaces to predefined meshes"""
# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging
import numpy as np

from capytaine.meshes.predefined import mesh_disk, mesh_vertical_cylinder, mesh_horizontal_cylinder
from capytaine.meshes.meshes import Mesh
from capytaine.bodies.bodies import FloatingBody

LOG = logging.getLogger(__name__)


##########
#  Disk  #
##########

class Disk(FloatingBody):
    """(One-sided) disk.
    Deprecated: please prefer capytaine.meshes.predefined.mesh_disk()

    Parameters
    ----------
    radius : float, optional
        radius of the disk
    resolution : 2-ple of int, optional
        number of panels along a radius and around the disk
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the disk
    normal: 3-ple of floats, optional
        normal vector, default: along x axis
    axial_symmetry : bool, optional
        if True, use the axial symmetry to speed up the computations
    reflection_symmetry : bool, optional
        if True, use the reflection symmetry to speed up the computations
    name : str, optional
        a string naming the floating body
    """

    def __init__(self, radius=1.0, resolution=(3, 5),
                 center=(0, 0, 0), normal=(1, 0, 0),
                 reflection_symmetry=False, axial_symmetry=False,
                 name=None):
        LOG.warning("Deprecation warning: The class Disk() is deprecated. "
                "Please prefer the function capytaine.meshes.predefined.mesh_disk()")

        if name is None:
            name = f"disk_{next(Mesh._ids)}"

        self.radius = float(radius)
        self.geometric_center = np.asarray(center, dtype=float)
        mesh = mesh_disk(radius=radius, center=center, normal=normal, resolution=resolution,
                reflection_symmetry=reflection_symmetry, axial_symmetry=axial_symmetry, name=f"{name}_mesh")
        FloatingBody.__init__(self, mesh=mesh, name=name)


##############
#  Cylinder  #
##############

class HorizontalCylinder(FloatingBody):
    """Horizontal cylinder
    Deprecated: please prefer capytaine.meshes.predefined.mesh_horizontal_cylinder()

    Parameters
    ----------
    length : float, optional
        length of the cylinder
    radius : float, optional
        radius of the cylinder
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the cylinder
    nx : int, optional
        number of circular slices
    ntheta : int, optional
        number of panels along a circular slice of the cylinder
    nr : int, optional
        number of panels along a radius on the extremities of the cylinder
    reflection_symmetry : bool, optional
        if True, returns a ReflectionSymmetricMesh
    translation_symmetry : bool, optional
        if True, uses a TranslationalSymmetricMesh internally for the main part of the cylinder
    name : str, optional
        a string naming the floating body
    """

    def __init__(self, length=10.0, radius=1.0, center=(0, 0, 0),
                 nx=10, ntheta=10, nr=2,
                 reflection_symmetry=True, translation_symmetry=False,
                 clever=None,
                 name=None):

        LOG.warning("Deprecation warning: The class HorizontalCylinder() is deprecated. "
                "Please prefer the function capytaine.meshes.predefined.mesh_horizontal_cylinder()")

        self.length = length
        self.radius = radius
        self.geometric_center = np.asarray(center, dtype=float)

        if name is None:
            name = f"cylinder_{next(Mesh._ids)}"

        mesh = mesh_horizontal_cylinder(length=length, radius=radius, center=center,
                resolution=(nr, ntheta, nx), reflection_symmetry=reflection_symmetry,
                translation_symmetry=translation_symmetry, name=f"{name}_mesh")
        FloatingBody.__init__(self, mesh=mesh, name=name)


class VerticalCylinder(FloatingBody):
    """Vertical cylinder.
    Deprecated: please prefer capytaine.meshes.predefined.mesh_vertical_cylinder()

    Parameters
    ----------
    length : float, optional
        length of the cylinder
    radius : float, optional
        radius of the cylinder
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the cylinder
    nx : int, optional
        number of circular slices
    ntheta : int, optional
        number of panels along a circular slice of the cylinder
    nr : int, optional
        number of panels along a radius on the extremities of the cylinder
    clever : bool, optional
        if True, uses the mesh symmetries
    name : str, optional
        a string naming the floating body
    """

    def __init__(self, length=10.0, radius=1.0, center=(0, 0, 0),
                 nx=10, ntheta=10, nr=2,
                 clever=True, name=None):
        LOG.warning("Deprecation warning: The class VerticalCylinder() is deprecated. "
                "Please prefer the function capytaine.meshes.predefined.mesh_vertical_cylinder()")

        self.length = length
        self.radius = radius
        self.geometric_center = np.asarray(center, dtype=float)

        if name is None:
            name = f"cylinder_{next(Mesh._ids)}"

        mesh = mesh_vertical_cylinder(length=length, radius=radius, center=center,
                resolution=(nr, ntheta, nx), reflection_symmetry=False,
                axial_symmetry=clever, name=f"{name}_mesh")

        FloatingBody.__init__(self, mesh=mesh, name=name)



#!/usr/bin/env python
# coding: utf-8

import capytaine as cpt

# Define the first body
sphere = cpt.FloatingBody(
        cpt.mesh_sphere(radius=1.0, center=(0, 0, -2.0),
                        resolution=(20, 20)),
        name="sphere_1")
sphere.add_translation_dof(name="Surge")
sphere.add_translation_dof(name="Heave")

# Define the second body
other_sphere = cpt.FloatingBody(
        cpt.mesh_sphere(radius=0.5, center=(-2, -3, -1),
                        resolution=(20, 20)),
        name="sphere_2")
other_sphere.add_translation_dof(name="Surge")
other_sphere.add_translation_dof(name="Heave")

# Define the third body
cylinder = cpt.FloatingBody(
        cpt.mesh_horizontal_cylinder(
            length=5.0, radius=1.0,
            center=(1.5, 3.0, -3.0),
            resolution=(20, 20, 3)),
        name="cylinder")
cylinder.add_translation_dof(name="Surge")
cylinder.add_translation_dof(name="Heave")

# Combine the three individual bodies into a single body.
all_bodies = cylinder + sphere + other_sphere

print("Merged body name:", all_bodies.name)
print("Merged body dofs:", list(all_bodies.dofs.keys()))

# The merged body can be used to define the problems in the usual way
problems = [cpt.RadiationProblem(body=all_bodies, radiating_dof=dof, omega=1.0) for dof in all_bodies.dofs]
problems += [cpt.DiffractionProblem(body=all_bodies, wave_direction=0.0, omega=1.0)]

# Solves the problem
solver = cpt.BEMSolver()
results = solver.solve_all(problems)
data = cpt.assemble_dataset(results)

print(data)



import logging
from rich.logging import RichHandler

def set_logging(level="INFO"):
    logging.basicConfig(level=level, format="%(message)s", handlers=[RichHandler(level=level, log_time_format="[%X]", show_path=False)], force=True)



"""3D display of floating body with VTK."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from capytaine.tools.optional_imports import import_optional_dependency
from capytaine.ui.vtk.mesh_viewer import MeshViewer

vtk = import_optional_dependency("vtk")

class FloatingBodyViewer(MeshViewer):

    def __init__(self):
        super().__init__()
        self.dofs_data = {}

    def add_body(self, body, **kwargs):
        self.add_mesh(body.mesh, **kwargs)

        for dof in body.dofs:
            vtk_data_array = vtk.vtkFloatArray()
            vtk_data_array.SetNumberOfComponents(3)
            vtk_data_array.SetNumberOfTuples(body.mesh.nb_faces)
            for i, vector in enumerate(body.dofs[dof]):
                vtk_data_array.SetTuple3(i, *vector)
            self.dofs_data[dof] = vtk_data_array
            # vtk_polydata.GetCellData().SetVectors(vtk_data_array)



import pytest

import numpy as np
import capytaine as cpt
from capytaine.bem.airy_waves import froude_krylov_force


def test_Froude_Krylov():
    sphere = cpt.FloatingBody(cpt.mesh_sphere(radius=1.0, resolution=(6, 12)).immersed_part())
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")

    problem = cpt.DiffractionProblem(body=sphere, omega=1.0, water_depth=np.inf)
    assert np.isclose(froude_krylov_force(problem)['Heave'], 27596, rtol=1e-3)

    problem = cpt.DiffractionProblem(body=sphere, omega=2.0, water_depth=np.inf)
    assert np.isclose(froude_krylov_force(problem)['Heave'], 22491, rtol=1e-3)

    problem = cpt.DiffractionProblem(body=sphere, omega=1.0, water_depth=10.0)
    assert np.isclose(froude_krylov_force(problem)['Heave'], 27610, rtol=1e-3)



# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from .__about__ import (
    __title__, __description__, __version__, __author__, __uri__, __license__
)

from capytaine.meshes.geometry import Axis, Plane, xOz_Plane, yOz_Plane, xOy_Plane
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import ReflectionSymmetricMesh, TranslationalSymmetricMesh, AxialSymmetricMesh
from capytaine.meshes.predefined.cylinders import mesh_disk, mesh_horizontal_cylinder, mesh_vertical_cylinder
from capytaine.meshes.predefined.spheres import mesh_sphere
from capytaine.meshes.predefined.rectangles import mesh_rectangle, mesh_parallelepiped

from capytaine.bodies.bodies import FloatingBody
from capytaine.bodies.dofs import rigid_body_dofs

from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import VerticalCylinder, HorizontalCylinder, Disk
from capytaine.bodies.predefined.rectangles import Rectangle, RectangularParallelepiped, OpenRectangularParallelepiped

from capytaine.bem.problems_and_results import RadiationProblem, DiffractionProblem
from capytaine.bem.solver import BEMSolver
from capytaine.bem.engines import BasicMatrixEngine, HierarchicalToeplitzMatrixEngine, HierarchicalPrecondMatrixEngine
from capytaine.green_functions.delhommeau import Delhommeau, XieDelhommeau

from capytaine.post_pro.free_surfaces import FreeSurface

from capytaine.io.mesh_loaders import load_mesh
from capytaine.io.xarray import assemble_dataset

from capytaine.ui.rich import set_logging

set_logging(level="WARNING")



"""Generate spherical bodies."""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>
import logging

import numpy as np
from numpy import pi

from capytaine.meshes.geometry import Axis
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.symmetric import AxialSymmetricMesh

LOG = logging.getLogger(__name__)


def mesh_sphere(*, radius=1.0, center=(0.0, 0.0, 0.0),
        resolution=(10, 10), faces_max_radius=None,
        axial_symmetry=False, name=None):
    """Sphere

    Parameters
    ----------
    radius : float
        radius of the sphere
    center : 3-ple or array of shape (3,)
        position of the geometric center of the sphere
    resolution : couple of ints
        number of panels along a meridian (or number of parallels-1) and
        along a parallel (or number of meridians-1)
    faces_max_radius : float, optional
        maximal radius of a panel. (Default: no maximal radius.)
        If the provided resolution is too coarse, the number of panels is
        changed to fit the constraint on the maximal radius.
    axial_symmetry : bool
        if True, use the axial symmetry to build the mesh (default: False)
    name : string
        a name identifying the sphere (default: "sphere_id" where id is an unique integer).
    """

    if name is None:
        name = f"sphere_{next(Mesh._ids)}"

    ntheta, nphi = resolution
    if faces_max_radius is not None:
        perimeter = 2*np.pi*radius
        estimated_max_radius = np.hypot(perimeter/ntheta, perimeter/nphi)/2
        if estimated_max_radius > faces_max_radius:
            ntheta = nphi = int(np.ceil(perimeter / (np.sqrt(2)*faces_max_radius)))

    theta = np.linspace(0.0, pi, ntheta+1)
    points_on_a_meridian = radius * np.stack([np.sin(theta), np.zeros_like(theta), -np.cos(theta)], axis=1)

    symmetry_axis = Axis(vector=[0, 0, 1], point=[0, 0, 0])
    mesh = AxialSymmetricMesh.from_profile(points_on_a_meridian, axis=symmetry_axis, nphi=nphi, name=name)

    if not axial_symmetry:
        mesh = mesh.merged()

    mesh.heal_mesh()
    mesh.translate(center)
    mesh.geometric_center = np.asarray(center, dtype=float)
    return mesh



import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import capytaine as cpt

cpt.set_logging('INFO')

# Generate body
mesh = cpt.mesh_horizontal_cylinder(
    length=3.0, radius=1.0, center=(0, 0, -1.01), resolution=(5, 30, 15)
)
body = cpt.FloatingBody(mesh)
body.add_translation_dof(name="Heave")

test_matrix = xr.Dataset(coords={
    'omega': np.linspace(0.5, 4, 40),
    'radiating_dof': list(body.dofs.keys()),
})

ds2 = cpt.BEMSolver(green_function=cpt.XieDelhommeau()).fill_dataset(test_matrix, body)
ds1 = cpt.BEMSolver(green_function=cpt.Delhommeau()).fill_dataset(test_matrix, body)

plt.figure()
ds1['added_mass'].plot(x='omega', label='Delhommeau')
ds2['added_mass'].plot(x='omega', label='XieDelhommeau')
plt.legend()

plt.figure()
ds1['radiation_damping'].plot(x='omega', label='Delhommeau')
ds2['radiation_damping'].plot(x='omega', label='XieDelhommeau')
plt.legend()

plt.show()



"""Quantitatively compare the results of Capytaine with the results from Nemoh 2."""

import pytest
import numpy as np
import capytaine as cpt

from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import HorizontalCylinder
from capytaine.post_pro.free_surfaces import FreeSurface
from capytaine.post_pro.kochin import compute_kochin


@pytest.fixture
def nemoh2_solver():
    gf = cpt.Delhommeau(
            tabulation_nr=328, tabulation_nz=46,
            tabulation_method='legacy', tabulation_nb_integration_points=251
            )
    solver = cpt.BEMSolver(
            engine=cpt.BasicMatrixEngine(matrix_cache_size=0),
            green_function=gf
            )
    return solver


def test_immersed_sphere(nemoh2_solver):
    """Compare with Nemoh 2.0 for a sphere in infinite fluid.

    The test is ran for two degrees of freedom; due to the symmetries of the problem, the results should be the same.
    They are actually slightly different due to the meshing of the sphere.
    """
    sphere = Sphere(radius=1.0, ntheta=10, nphi=40, clip_free_surface=False)
    sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")

    problem = cpt.RadiationProblem(body=sphere, radiating_dof="Heave", free_surface=np.inf, water_depth=np.inf)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.added_masses["Heave"],       2187, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.added_masses["Surge"],        0.0, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"],  0.0, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Surge"],  0.0, atol=1e-3*sphere.volume*problem.rho)

    problem = cpt.RadiationProblem(body=sphere, radiating_dof="Surge", free_surface=np.inf, water_depth=np.inf)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.added_masses["Surge"],       2194, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.added_masses["Heave"],        0.0, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Surge"],  0.0, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"],  0.0, atol=1e-3*sphere.volume*problem.rho)


def test_build_matrix_of_rankine_and_reflected_rankine(nemoh2_solver):
    gf = nemoh2_solver.green_function
    sphere = Sphere(radius=1.0, ntheta=2, nphi=3, clip_free_surface=True)

    S, V = gf.evaluate(sphere.mesh, sphere.mesh, 0.0, np.inf, 0.0)
    S_ref = np.array([[-0.15413386, -0.21852682, -0.06509213, -0.16718431, -0.06509213, -0.16718431],
                      [-0.05898834, -0.39245688, -0.04606661, -0.18264734, -0.04606661, -0.18264734],
                      [-0.06509213, -0.16718431, -0.15413386, -0.21852682, -0.06509213, -0.16718431],
                      [-0.04606661, -0.18264734, -0.05898834, -0.39245688, -0.04606661, -0.18264734],
                      [-0.06509213, -0.16718431, -0.06509213, -0.16718431, -0.15413386, -0.21852682],
                      [-0.04606661, -0.18264734, -0.04606661, -0.18264734, -0.05898834, -0.39245688]])
    assert np.allclose(S, S_ref)

    S, V = gf.evaluate(sphere.mesh, sphere.mesh, 0.0, np.inf, np.inf)
    S_ref = np.array([[-0.12666269, -0.07804937, -0.03845837, -0.03993999, -0.03845837, -0.03993999],
                      [-0.02106031, -0.16464793, -0.01169102, -0.02315146, -0.01169102, -0.02315146],
                      [-0.03845837, -0.03993999, -0.12666269, -0.07804937, -0.03845837, -0.03993999],
                      [-0.01169102, -0.02315146, -0.02106031, -0.16464793, -0.01169102, -0.02315146],
                      [-0.03845837, -0.03993999, -0.03845837, -0.03993999, -0.12666269, -0.07804937],
                      [-0.01169102, -0.02315146, -0.01169102, -0.02315146, -0.02106031, -0.16464793]])
    assert np.allclose(S, S_ref)


def test_floating_sphere_finite_freq(nemoh2_solver):
    """Compare with Nemoh 2.0 for some cases of a heaving sphere at the free surface in infinite depth."""
    sphere = Sphere(radius=1.0, ntheta=3, nphi=12, clip_free_surface=True)
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")

    # omega = 1, radiation
    problem = cpt.RadiationProblem(body=sphere, omega=1.0, water_depth=np.inf)
    result = nemoh2_solver.solve(problem, keep_details=True)
    assert np.isclose(result.added_masses["Heave"],       1819.6, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"], 379.39, atol=1e-3*sphere.volume*problem.rho)

    # omega = 1, free surface
    grid = np.meshgrid(np.linspace(-50.0, 50.0, 5), np.linspace(-50.0, 50.0, 5))
    eta = nemoh2_solver.compute_free_surface_elevation(grid, result)
    ref = np.array(
            [[-0.4340802E-02-0.4742809E-03j, -0.7986111E-03+0.4840984E-02j, 0.2214827E-02+0.4700642E-02j, -0.7986111E-03+0.4840984E-02j, -0.4340803E-02-0.4742807E-03j],
             [-0.7986111E-03+0.4840984E-02j, 0.5733187E-02-0.2179381E-02j, 0.9460892E-03-0.7079404E-02j, 0.5733186E-02-0.2179381E-02j, -0.7986110E-03+0.4840984E-02j],
             [0.2214827E-02+0.4700643E-02j, 0.9460892E-03-0.7079403E-02j, -0.1381670E-01+0.6039315E-01j, 0.9460892E-03-0.7079405E-02j, 0.2214827E-02+0.4700643E-02j],
             [-0.7986111E-03+0.4840984E-02j, 0.5733186E-02-0.2179381E-02j, 0.9460891E-03-0.7079404E-02j, 0.5733187E-02-0.2179380E-02j, -0.7986113E-03+0.4840984E-02j],
             [-0.4340803E-02-0.4742807E-03j, -0.7986111E-03+0.4840984E-02j, 0.2214827E-02+0.4700643E-02j, -0.7986113E-03+0.4840983E-02j, -0.4340803E-02-0.4742809E-03j]]
        )
    assert np.allclose(eta/(-1j*problem.omega), ref, rtol=1e-4)

    # omega = 1, diffraction
    problem = cpt.DiffractionProblem(body=sphere, omega=1.0, water_depth=np.inf)
    result = nemoh2_solver.solve(problem, keep_details=True)
    assert np.isclose(result.forces["Heave"], 1834.9 * np.exp(-2.933j), rtol=1e-3)

    # omega = 1, Kochin function of diffraction problem

    kochin = compute_kochin(result, np.linspace(0, np.pi, 10))

    ref_kochin = np.array([
        0.20229*np.exp(-1.5872j), 0.20369*np.exp(-1.5871j),
        0.20767*np.exp(-1.5868j), 0.21382*np.exp(-1.5863j),
        0.22132*np.exp(-1.5857j), 0.22931*np.exp(-1.5852j),
        0.23680*np.exp(-1.5847j), 0.24291*np.exp(-1.5843j),
        0.24688*np.exp(-1.5841j), 0.24825*np.exp(-1.5840j),
        ])

    assert np.allclose(kochin, ref_kochin, rtol=1e-3)

    # omega = 2, radiation
    problem = cpt.RadiationProblem(body=sphere, omega=2.0, water_depth=np.inf)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.added_masses["Heave"],       1369.3, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"], 1425.6, atol=1e-3*sphere.volume*problem.rho)

    # omega = 2, diffraction
    problem = cpt.DiffractionProblem(body=sphere, omega=2.0, water_depth=np.inf)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.forces["Heave"], 5846.6 * np.exp(-2.623j), rtol=1e-3)


def test_alien_sphere(nemoh2_solver):
    """Compare with Nemoh 2.0 for some cases of a heaving sphere at the free surface in infinite depth
    for a non-usual gravity and density."""
    sphere = Sphere(radius=1.0, ntheta=3, nphi=12, clip_free_surface=True)
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")

    # radiation
    problem = cpt.RadiationProblem(body=sphere, rho=450.0, g=1.625, omega=1.0, radiating_dof="Heave")
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.added_masses["Heave"],       515, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"], 309, atol=1e-3*sphere.volume*problem.rho)

    # diffraction
    problem = cpt.DiffractionProblem(body=sphere, rho=450.0, g=1.625, omega=1.0, water_depth=np.inf)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.forces["Heave"], 548.5 * np.exp(-2.521j), rtol=1e-2)


def test_floating_sphere_finite_depth(nemoh2_solver):
    """Compare with Nemoh 2.0 for some cases of a heaving sphere at the free surface in finite depth."""
    sphere = Sphere(radius=1.0, ntheta=3, nphi=12, clip_free_surface=True)
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")

    # omega = 1, radiation
    problem = cpt.RadiationProblem(body=sphere, omega=1.0, radiating_dof="Heave", water_depth=10.0)
    result = nemoh2_solver.solve(problem, keep_details=True)
    assert np.isclose(result.added_masses["Heave"],       1740.6, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"], 380.46, atol=1e-3*sphere.volume*problem.rho)

    kochin = compute_kochin(result, np.linspace(0, np.pi, 3))
    assert np.allclose(kochin, np.roll(kochin, 1))  # The far field is the same in all directions.
    assert np.isclose(kochin[0]/(-1j*problem.omega), -0.2267+3.49e-3j, rtol=1e-3)

    # omega = 1, diffraction
    problem = cpt.DiffractionProblem(body=sphere, omega=1.0, wave_direction=0.0, water_depth=10.0)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.forces["Heave"], 1749.4 * np.exp(-2.922j), rtol=1e-3)

    # omega = 2, radiation
    problem = cpt.RadiationProblem(body=sphere, omega=2.0, radiating_dof="Heave", water_depth=10.0)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.added_masses["Heave"],       1375.0, atol=1e-3*sphere.volume*problem.rho)
    assert np.isclose(result.radiation_dampings["Heave"], 1418.0, atol=1e-3*sphere.volume*problem.rho)

    # omega = 2, diffraction
    problem = cpt.DiffractionProblem(body=sphere, omega=2.0, wave_direction=0.0, water_depth=10.0)
    result = nemoh2_solver.solve(problem)
    assert np.isclose(result.forces["Heave"], 5872.8 * np.exp(-2.627j), rtol=1e-3)


def test_two_distant_spheres_in_finite_depth(nemoh2_solver):
    radius = 0.5
    resolution = 4
    perimeter = 2*np.pi*radius
    buoy = Sphere(radius=radius, center=(0.0, 0.0, 0.0),
                  ntheta=int(perimeter*resolution/2), nphi=int(perimeter*resolution),
                  clip_free_surface=True, axial_symmetry=False, name="buoy")
    other_buoy = buoy.translated_x(20, name="other_buoy")
    both_buoys = buoy.join_bodies(other_buoy)
    both_buoys.add_translation_dof(name="Surge")
    problem = cpt.RadiationProblem(body=both_buoys, radiating_dof="Surge", water_depth=10, omega=7.0)
    result = nemoh2_solver.solve(problem)

    total_volume = 2*4/3*np.pi*radius**3
    assert np.isclose(result.added_masses['Surge'], 124.0, atol=1e-3*total_volume*problem.rho)
    assert np.isclose(result.radiation_dampings['Surge'], 913.3, atol=1e-3*total_volume*problem.rho)


def test_multibody(nemoh2_solver):
    """Compare with Nemoh 2.0 for two bodies."""
    sphere = Sphere(radius=1.0, ntheta=5, nphi=20)
    sphere.translate_z(-2.0)
    sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")

    cylinder = HorizontalCylinder(length=5.0, radius=1.0,
                                  nx=10, nr=1, ntheta=10)
    cylinder.translate([+1.5, 3.0, -3.0])
    cylinder.add_translation_dof(direction=(1, 0, 0), name="Surge")
    cylinder.add_translation_dof(direction=(0, 0, 1), name="Heave")

    both = cylinder + sphere
    total_volume = cylinder.volume + sphere.volume
    # both.show()

    problems = [cpt.RadiationProblem(body=both, radiating_dof=dof, omega=1.0) for dof in both.dofs]
    problems += [cpt.DiffractionProblem(body=both, wave_direction=0.0, omega=1.0)]
    results = [nemoh2_solver.solve(problem) for problem in problems]
    data = cpt.assemble_dataset(results)

    data_from_nemoh_2 = np.array([
        [3961.86548, 50.0367661, -3.32347107, 6.36901855E-02, 172.704819, 19.2018471, -5.67303181, -2.98873377],
        [-3.08301544, 5.72392941E-02, 14522.1689, 271.796814, 128.413834, 6.03351116, 427.167358, 64.1587067],
        [161.125534, 17.8332844, 126.392113, 5.88006783, 2242.47412, 7.17850924, 1.29002571, 0.393169671],
        [-5.02560759, -2.75930357, 419.927460, 63.3179016, 1.23501396, 0.416424811, 2341.57593, 15.8266096],
    ])

    dofs_names = list(both.dofs.keys())
    assert np.allclose(
        data['added_mass'].sel(omega=1.0, radiating_dof=dofs_names, influenced_dof=dofs_names).values,
        data_from_nemoh_2[:, ::2],
        atol=1e-3*total_volume*problems[0].rho
    )
    assert np.allclose(
        data['radiation_damping'].sel(omega=1.0, radiating_dof=dofs_names, influenced_dof=dofs_names).values,
        data_from_nemoh_2[:, 1::2],
        atol=1e-3*total_volume*problems[0].rho
    )



"""Generate spherical bodies."""
# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging

import numpy as np

from capytaine.meshes import Mesh
from capytaine.meshes.predefined import mesh_sphere
from capytaine.bodies.bodies import FloatingBody

LOG = logging.getLogger(__name__)


class Sphere(FloatingBody):
    """Sphere
    Deprecated: please prefer capytaine.meshes.predefined.mesh_sphere()

    Parameters
    ----------
    radius : float
        radius of the sphere
    center : 3-ple or array of shape (3,)
        position of the geometric center of the sphere
    ntheta : int
        number of panels along a meridian (or number of parallels-1)
    nphi : int
        number of panels along a parallel (or number of meridians-1)
    axial_symmetry : bool
        if True, use the axial symmetry to build the mesh (default: True)
    clip_free_surface : bool
        if True, only mesh the part of the sphere where z < 0 (default: False),
        can be used with center to obtain any clipped sphere,
        if True, then ntheta is the number of parallel below the free surface.
    name : string
        a name identifying the sphere (default: "sphere_id" where id is an unique integer).
    """

    def __init__(self, *, radius=1.0, center=(0, 0, 0),
                 ntheta=10, nphi=10, clip_free_surface=False,
                 axial_symmetry=True, clever=None,
                 name=None):

        LOG.warning("Deprecation warning: The class Sphere() is deprecated. "
                "Please prefer the function capytaine.meshes.predefined.mesh_sphere()")

        if clever is not None:
            LOG.warning("Deprecation warning: `clever` argument for Sphere is deprecated. "
                        "Use `axial_symmetry` instead.")

        if name is None:
            name = f"sphere_{next(Mesh._ids)}"

        if clip_free_surface:
            if center[2] < -radius:  # fully immersed
                pass
            elif center[2] < radius:
                ntheta = int(ntheta*np.pi/np.arccos(center[2]/radius))
            else:
                raise ValueError("Impossible to mesh the immersed hull of a sphere completely out of the water")

        mesh = mesh_sphere(radius=radius, center=center, resolution=(ntheta, nphi), axial_symmetry=axial_symmetry, name=f"{name}_mesh")

        if clip_free_surface:
            mesh.keep_immersed_part()

        self.radius = radius
        self.geometric_center = np.array(center, dtype=float)
        FloatingBody.__init__(self, mesh=mesh, name=name)



import numpy as np
from capytaine.bodies import FloatingBody
from capytaine.post_pro.free_surfaces import FreeSurface
from capytaine.meshes import Mesh, CollectionOfMeshes


def _normalize_points(points, keep_mesh=False):
    if isinstance(points, (FloatingBody, FreeSurface)):
        if keep_mesh:
            return points.mesh, (points.mesh.nb_faces,)
        else:
            return points.mesh.faces_centers, (points.mesh.nb_faces,)

    if isinstance(points, (Mesh, CollectionOfMeshes)):
        if keep_mesh:
            return points, (points.nb_faces,)
        else:
            return points.faces_centers, (points.nb_faces,)

    points = np.asarray(points)

    if points.ndim == 1:  # A single point has been provided
        output_shape = (1,)
        points = points.reshape((1, points.shape[0]))

    elif points.ndim == 2:
        output_shape = (points.shape[0],)

    elif points.ndim > 2:
        # `points` is expected to be the results of a meshgrid. Points has shape (d, nx, ny, ...)
        output_shape = points.shape[1:]
        points = points.reshape(points.shape[0], -1).transpose()
        # points is now a (nx*ny*... , d) array

    else:
        raise ValueError(f"Expected a list of points or a mesh, but got instead: {points}")

    return points, output_shape

def _normalize_free_surface_points(points, keep_mesh=False):
    if keep_mesh and isinstance(points, (FloatingBody, FreeSurface)):
        return points.mesh, (points.mesh.nb_faces,)

    if keep_mesh and isinstance(points, (Mesh, CollectionOfMeshes)):
        return points, (points.nb_faces,)

    points, output_shape = _normalize_points(points, keep_mesh)

    if points.ndim == 2 and points.shape[1] == 2:  # Only x and y have been provided
        points = np.concatenate([points, np.zeros((points.shape[0], 1))], axis=1)

    return points, output_shape



#!/usr/bin/env python
# coding: utf-8
"""Test related to the computation of Kochin functions."""

import numpy as np
from numpy import pi
import pandas as pd
import xarray as xr
import pytest

import capytaine as cpt
from capytaine.io.xarray import kochin_data_array


def test_kochin_array_diffraction():
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb_diff = cpt.DiffractionProblem(body=body, wavelength=2.0, wave_direction=0.0)
    res_diff = solver.solve(pb_diff)
    kds_diff = kochin_data_array([res_diff], np.linspace(0.0, np.pi, 3))
    assert "kochin_diffraction" in kds_diff
    assert "kochin" not in kds_diff
    assert len(kds_diff.dims) == 3

def test_kochin_array_radiation():
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb_rad = cpt.RadiationProblem(body=body, wavelength=2.0, radiating_dof="Heave")
    res_rad = solver.solve(pb_rad)
    kds_rad = kochin_data_array([res_rad], np.linspace(0.0, np.pi, 3))
    assert "kochin_diffraction" not in kds_rad
    assert "kochin" in kds_rad
    assert len(kds_rad.dims) == 3

def test_kochin_array_both_radiation_and_diffraction():
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb_diff = cpt.DiffractionProblem(body=body, wavelength=2.0, wave_direction=0.0)
    pb_rad = cpt.RadiationProblem(body=body, wavelength=2.0, radiating_dof="Heave")
    res_both = solver.solve_all([pb_rad, pb_diff])
    kds_both = kochin_data_array(res_both, np.linspace(0.0, np.pi, 3))
    assert "kochin_diffraction" in kds_both
    assert "kochin" in kds_both
    assert len(kds_both.dims) == 4

def test_kochin_array_diffraction_with_forward_speed():
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb_diff = [cpt.DiffractionProblem(body=body, wavelength=2.0, wave_direction=0.0, forward_speed=u) for u in [0.0, 1.0]]
    res_diff = solver.solve_all(pb_diff)
    kds_diff = kochin_data_array(res_diff, np.linspace(0.0, np.pi, 3))
    assert "kochin_diffraction" in kds_diff
    assert "kochin" not in kds_diff
    assert len(kds_diff.dims) == 4

def test_kochin_array_radiation_with_forward_speed():
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb_rad = [cpt.RadiationProblem(body=body, wavelength=2.0, radiating_dof="Heave", forward_speed=u, wave_direction=pi) for u in [0.0, 1.0]]
    res_rad = solver.solve_all(pb_rad)
    kds_rad = kochin_data_array(res_rad, np.linspace(0.0, np.pi, 3))
    assert "kochin_diffraction" not in kds_rad
    assert "kochin" in kds_rad
    assert len(kds_rad.dims) == 4

def test_kochin_array_both_radiation_and_diffraction_with_forward_speed():
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb_diff = [cpt.DiffractionProblem(body=body, wavelength=2.0, wave_direction=pi, forward_speed=u) for u in [0.0, 1.0]]
    pb_rad = [cpt.RadiationProblem(body=body, wavelength=2.0, radiating_dof="Heave", wave_direction=pi, forward_speed=u) for u in [0.0, 1.0]]
    res_both = solver.solve_all(pb_rad + pb_diff)
    kds_both = kochin_data_array(res_both, np.linspace(0.0, np.pi, 3))
    assert "kochin_diffraction" in kds_both
    assert "kochin" in kds_both
    assert len(kds_both.dims) == 5

def test_fill_dataset_with_kochin_functions():
    sphere = cpt.Sphere(clip_free_surface=True)
    sphere.add_translation_dof(name="Heave")
    solver = cpt.BEMSolver()

    test_matrix = xr.Dataset(coords={
        'omega': [1.0],
        'theta': [0.0, pi/2],
        'radiating_dof': ["Heave"],
    })
    ds = solver.fill_dataset(test_matrix, [sphere])
    assert 'theta' in ds.coords
    assert 'kochin' in ds
    assert 'kochin_diffraction' not in ds

    # Because of the symmetries of the body
    assert np.isclose(ds['kochin'].sel(radiating_dof="Heave", theta=0.0),
                      ds['kochin'].sel(radiating_dof="Heave", theta=pi/2))

    test_matrix = xr.Dataset(coords={
        'omega': [1.0],
        'radiating_dof': ["Heave"],
        'wave_direction': [-pi/2, 0.0],
        'theta': [0.0, pi/2],
    })
    ds = solver.fill_dataset(test_matrix, [sphere])
    assert 'theta' in ds.coords
    assert 'kochin' in ds
    assert 'kochin_diffraction' in ds

    # Because of the symmetries of the body
    assert np.isclose(ds['kochin_diffraction'].sel(wave_direction=-pi/2, theta=0.0),
                      ds['kochin_diffraction'].sel(wave_direction=0.0, theta=pi/2))



import xarray as xr
import numpy as np
import capytaine as cpt
import pytest
import os

from capytaine.io.xarray import problems_from_dataset, separate_complex_values, merge_complex_values
from capytaine.io.legacy import export_hydrostatics


def test_incomplete_test_matrix():
    body = cpt.Sphere()
    test_matrix = xr.Dataset(coords={
        "omega": np.linspace(0, 1, 2),
        })
    with pytest.raises(ValueError):
        problems_from_dataset(test_matrix, body)


def test_remove_complex_values():
    original_dataset = xr.Dataset(
        data_vars={
            'variable1': (['x', 'y'], np.random.rand(4, 5)),
            'variable2': (['x', 'y', 'z'], np.random.rand(4, 5, 3) + 1j * np.random.rand(4, 5, 3)),
            'variable3': (['y', 'z'], np.random.rand(5, 3) + 1j * np.random.rand(5, 3)),
        },
        coords={
            'x': (['x'], np.linspace(0, 10, 4))
        })

    real_dataset = separate_complex_values(original_dataset)
    assert np.allclose(real_dataset['variable3'].sel(complex='re').data,
                       np.real(original_dataset['variable3'].data),
                       atol=0.1)
    assert set(original_dataset.dims) == set(real_dataset.dims) - {'complex'}

    complex_dataset = merge_complex_values(real_dataset)
    assert set(original_dataset.dims) == set(complex_dataset.dims)


def test_xarray_dataset_with_more_data():
    # Store some mesh data when several bodies in dataset
    bodies = [
        cpt.Sphere(radius=1, ntheta=3, nphi=3, name="sphere_1"),
        cpt.Sphere(radius=2, ntheta=5, nphi=3, name="sphere_2"),
        cpt.Sphere(radius=3, ntheta=7, nphi=3, name="sphere_3"),
    ]
    for body in bodies:
        body.keep_immersed_part()
        body.add_translation_dof(name="Heave")

    problems = [cpt.RadiationProblem(body=b, radiating_dof="Heave", omega=1.0) for b in bodies]
    results = cpt.BEMSolver().solve_all(problems)

    ds = cpt.assemble_dataset(results, mesh=True)
    assert 'nb_faces' in ds.coords
    assert set(ds.coords['nb_faces'].values) == set([b.mesh.nb_faces for b in bodies])


def test_dataset_from_bemio():
    bemio = pytest.importorskip("bemio.io.wamit", reason="Bemio not installed, test skipped.")
    current_file_path = os.path.dirname(os.path.abspath(__file__))
    out_file = os.path.join(current_file_path, "Bemio_verification_cases", "sphere.out")
    bemio_data = bemio.read(out_file)

    new_dataset = cpt.assemble_dataset(bemio_data)
    assert (np.moveaxis(bemio_data.body[0].am.all, 2, 0) * bemio_data.body[0].rho == \
        new_dataset['added_mass'].values).all()


def test_legacy_export_hydrostatics():
    # create two cylinders & compute inertia & hydrostatic data
    cylinder1_mesh = cpt.meshes.predefined.mesh_vertical_cylinder(length=5.0, center = (1.0, 1.0, -0.5))
    cylinder1 = cpt.FloatingBody(cylinder1_mesh, center_of_mass=[1.0, 1.0, -2.0])
    cylinder1.add_all_rigid_body_dofs()
    cylinder1.compute_rigid_body_inertia()
    cylinder1.compute_hydrostatics()

    cylinder2_mesh = cpt.meshes.predefined.mesh_vertical_cylinder(length=5.0, center = (5.0, 1.0, -0.5))
    cylinder2 = cpt.FloatingBody(cylinder2_mesh, center_of_mass=[5.0, 1.0, -2.0])
    cylinder2.add_all_rigid_body_dofs()
    cylinder2.compute_rigid_body_inertia()
    cylinder2.compute_hydrostatics()

    current_file_path = os.path.dirname(os.path.abspath(__file__))
    testing_directory = os.path.join(current_file_path, 'io_legacy_cases')

    # export hydrostatics
    export_hydrostatics(os.path.join(testing_directory, 'single_body'), cylinder1)
    export_hydrostatics(os.path.join(testing_directory, 'single_body_list'), [cylinder1])
    export_hydrostatics(os.path.join(testing_directory, 'two_bodies_list'), [cylinder1, cylinder2])


    # Check single body Hydrostatics.dat
    with open(os.path.join(testing_directory, "single_body/Hydrostatics.dat"), "r") as f:
        single_body_Hydrostatics = f.read()
    with open(os.path.join(testing_directory, "reference_data/single_body/Hydrostatics.dat"), "r") as f:
        single_body_Hydrostatics_ref = f.read()
    assert single_body_Hydrostatics == single_body_Hydrostatics_ref


    # Check single body (list) Hydrostatics.dat
    with open(os.path.join(testing_directory, "single_body_list/Hydrostatics.dat"), "r") as f:
        single_body_list_Hydrostatics = f.read()
    with open(os.path.join(testing_directory, "reference_data/single_body_list/Hydrostatics.dat"), "r") as f:
        single_body_list_Hydrostatics_ref = f.read()
    assert single_body_list_Hydrostatics == single_body_list_Hydrostatics_ref


    # Check two bodies (list) Hydrostatics_0.dat
    with open(os.path.join(testing_directory, "two_bodies_list/Hydrostatics_0.dat"), "r") as f:
        two_bodies_Hydrostatics_0 = f.read()
    with open(os.path.join(testing_directory, "reference_data/two_bodies_list/Hydrostatics_0.dat"), "r") as f:
        two_bodies_Hydrostatics_0_ref = f.read()
    assert two_bodies_Hydrostatics_0 == two_bodies_Hydrostatics_0_ref

    # Check two bodies (list) Hydrostatics_1.dat
    with open(os.path.join(testing_directory, "two_bodies_list/Hydrostatics_1.dat"), "r") as f:
        two_bodies_Hydrostatics_1 = f.read()
    with open(os.path.join(testing_directory, "reference_data/two_bodies_list/Hydrostatics_1.dat"), "r") as f:
        two_bodies_Hydrostatics_1_ref = f.read()
    assert two_bodies_Hydrostatics_1 == two_bodies_Hydrostatics_1_ref


    # Check single body KH.dat
    single_body_KH = np.loadtxt(os.path.join(testing_directory, "single_body/KH.dat"))
    single_body_KH_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/single_body/KH.dat"))
    np.testing.assert_allclose(single_body_KH, single_body_KH_ref, atol=1e-6)

    # Check single body (list) KH.dat
    single_body_list_KH = np.loadtxt(os.path.join(testing_directory, "single_body_list/KH.dat"))
    single_body_list_KH_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/single_body_list/KH.dat"))
    np.testing.assert_allclose(single_body_list_KH, single_body_list_KH_ref, atol=1e-6)

    # Check two bodies (list) KH_0.dat
    two_bodies_KH_0 = np.loadtxt(os.path.join(testing_directory, "two_bodies_list/KH_0.dat"))
    two_bodies_KH_0_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/two_bodies_list/KH_0.dat"))
    np.testing.assert_allclose(two_bodies_KH_0, two_bodies_KH_0_ref, atol=1e-6)

    # Check two bodies (list) KH_1.dat
    two_bodies_KH_1 = np.loadtxt(os.path.join(testing_directory, "two_bodies_list/KH_1.dat"))
    two_bodies_KH_1_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/two_bodies_list/KH_1.dat"))
    np.testing.assert_allclose(two_bodies_KH_1, two_bodies_KH_1_ref, atol=1e-6)



#!/usr/bin/env python

import numpy as np
import capytaine as cpt

# Load a mesh
original_mesh = cpt.mesh_sphere()

# METHOD 1
# Get the indices of the panels such that the center is in the half-space y > 0.
indices_of_some_faces = np.where(original_mesh.faces_centers[:, 1] > 0)[0]
half_mesh = original_mesh.extract_faces(indices_of_some_faces)

# METHOD 2
# Clip the mesh
half_mesh = original_mesh.clip(plane=cpt.xOz_Plane)

# Define a new body by reflecting the half mesh across the xOz plane.
body = cpt.FloatingBody(
    cpt.ReflectionSymmetricMesh(half_mesh, cpt.xOz_Plane),
    name="symmetric_body")

# Display the body with VTK.
body.show()



"""Generate rectangular bodies."""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>
import logging
from itertools import product

import numpy as np

from capytaine.meshes.geometry import xOz_Plane, yOz_Plane
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.symmetric import TranslationalSymmetricMesh, ReflectionSymmetricMesh
from capytaine.meshes.collections import CollectionOfMeshes

LOG = logging.getLogger(__name__)

def mesh_rectangle(*, size=(5.0, 5.0), center=(0.0, 0.0, 0.0),
        resolution=(5, 5), faces_max_radius=None,
        normal=(0.0, 0.0, 1.0),
        translation_symmetry=False, reflection_symmetry=False,
        name=None):
    """One-sided rectangle.

    By default, the rectangle is horizontal, the normals are oriented upwards.

    Parameters
    ----------
    size : couple of floats, optional
        dimensions of the rectangle (width and height)
    center : 3-ple of floats, optional
        position of the geometric center of the rectangle, default: (0, 0, 0)
    resolution : couple of ints, optional
        number of faces along each of the two directions
    faces_max_radius : float, optional
        maximal radius of a panel. (Default: no maximal radius.)
        If the provided resolution is too coarse, the number of panels is
        changed to fit the constraint on the maximal radius.
    normal: 3-ple of floats, optional
        normal vector, default: (0, 0, 1)
    translation_symmetry : bool, optional
        if True, use the translation symmetry to speed up the computations
    reflection_symmetry : bool, optional
        if True, use the reflection symmetry to speed up the computations
    name : string, optional
        a name for the body
    """

    assert len(size) == 2, "Size of a rectangle should be given as a couple of values."
    assert all([h > 0 for h in size]), "Size of the rectangle mesh should be given as positive values."

    assert len(resolution) == 2, "Resolution of a rectangle should be given as a couple a values."
    assert all([h > 0 for h in resolution]), "Resolution of the rectangle mesh should be given as positive values."
    assert all([i == int(i) for i in resolution]), "Resolution of a rectangle should be given as integer values."

    assert len(center) == 3, "Position of the center of a rectangle should be given a 3-ple of values."

    width, height = size
    nw, nh = resolution

    if faces_max_radius is not None:
        estimated_max_radius = np.hypot(width/nw, height/nh)/2
        if estimated_max_radius > faces_max_radius:
            nw = int(np.ceil(width / (np.sqrt(2) * faces_max_radius)))
            nh = int(np.ceil(height / (np.sqrt(2) * faces_max_radius)))

    if name is None:
        name = f"rectangle_{next(Mesh._ids)}"

    if translation_symmetry and reflection_symmetry:
        raise NotImplementedError("Rectangle generation with both reflection and translation symmetries "
                                  "has not been implemented.")

    if reflection_symmetry:
        if nw % 2 == 1:
            raise ValueError("To use the reflection symmetry of the mesh, "
                             "it should have an even number of panels in this direction.")

        half_mesh = mesh_rectangle(size=(width/2, height), resolution=(nw//2, nh),
                                   center=(0, -width/4, 0), normal=(0.0, 0.0, 1.0),
                                   translation_symmetry=False, reflection_symmetry=False,
                                   name=f"half_of_{name}")
        mesh = ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=name)

    elif translation_symmetry:
            strip = mesh_rectangle(size=(width/nw, height), resolution=(1, nh),
                                   center=(0, -width/2 + width/(2*nw), 0), normal=(0.0, 0.0, 1.0),
                                   translation_symmetry=False, reflection_symmetry=False,
                                   name=f"strip_of_{name}")
            mesh = TranslationalSymmetricMesh(strip,
                                              translation=np.asarray([0, width/nw, 0]),
                                              nb_repetitions=int(nw)-1,
                                              name=name)

    else:
        y_range = np.linspace(-width/2, width/2, nw+1)
        z_range = np.linspace(-height/2, height/2, nh+1)
        nodes = np.array(list(product([0.0], y_range, z_range)), dtype=float)
        panels = np.array([(j+i*(nh+1), j+1+i*(nh+1), j+1+(i+1)*(nh+1), j+(i+1)*(nh+1))
                             for (i, j) in product(range(nw), range(nh))])

        mesh = Mesh(nodes, panels, name=name)

    mesh.heal_mesh()
    mesh.translate(center)
    mesh.rotate_around_center_to_align_vectors(center, mesh.faces_normals[0], normal)
    mesh.geometric_center = np.asarray(center, dtype=float)
    return mesh


def mesh_parallelepiped(size=(1.0, 1.0, 1.0), center=(0, 0, 0),
                        resolution=(4, 4, 4), faces_max_radius=None,
                        missing_sides=set(), reflection_symmetry=False, translation_symmetry=False,
                        name=None):
    """Six rectangles forming a parallelepiped.

    Parameters
    ----------
    size : 3-ple of floats, optional
        dimensions of the parallelepiped (width, thickness, height) for coordinates (x, y, z).
    center : 3-ple of floats, optional
        coordinates of the geometric center of the parallelepiped
    resolution : 3-ple of ints, optional
        number of faces along the three directions
    faces_max_radius : float, optional
        maximal radius of a panel. (Default: no maximal radius.)
        If the provided resolution is too coarse, the number of panels is
        changed to fit the constraint on the maximal radius.
    missing_sides : set of string, optional
        if one of the keyword "top", "bottom", "front", "back", "left", "right" is in the set,
        then the corresponding side is not included in the parallelepiped.
        May be ignored when building a mesh with a symmetry.
    reflection_symmetry : bool, optional
        use xOz and yOz symmetry plane to generate the mesh
    translation_symmetry : bool, optional
        if True, use the translation symmetry in the x direction to speed up the computations.
        To use the translation symmetry in the y direction, create a x-symmetric body and then rotate it by pi/2.
    name : string, optional
        a name for the body
    """

    assert len(size) == 3, "Size of a rectangular parallelepiped should be given as a 3-ple of values."
    assert all([h > 0 for h in size]), "Size of the rectangular mesh should be given as positive values."

    assert len(resolution) == 3, "Resolution of a rectangular parallelepiped should be given as a 3-ple a values."
    assert all([h > 0 for h in resolution]), "Resolution of the rectangular parallelepiped mesh " \
                                             "should be given as positive values."
    assert all([i == int(i) for i in resolution]), "Resolution of a rectangular parallelepiped " \
                                                   "should be given as integer values."

    assert len(center) == 3, "Position of the center of a parallelepiped should be given a 3-ple of values."

    width, thickness, height = size
    nw, nth, nh = resolution

    if faces_max_radius is not None:
        dw, dh, dth = width/nw, height/nh, thickness/nth
        estimated_max_radius = max(
                np.hypot(dw, dh)/2,
                np.hypot(dw, dth)/2,
                np.hypot(dth, dh)/2,
                )
        if estimated_max_radius > faces_max_radius:
            nw = int(np.ceil(width / (np.sqrt(2) * faces_max_radius)))
            nth = int(np.ceil(thickness / (np.sqrt(2) * faces_max_radius)))
            nh = int(np.ceil(height / (np.sqrt(2) * faces_max_radius)))

    if name is None:
        name = f"rectangular_parallelepiped_{next(Mesh._ids)}"

    if translation_symmetry and reflection_symmetry:
        raise NotImplementedError("Parallelepiped generation with both reflection and translation symmetries "
                                  "has not been implemented.")

    if reflection_symmetry:
        if (nw % 2 == 1 or nth % 2 == 1):
            raise ValueError("To use the reflection symmetry of the mesh, "
                             "it should have an even number of panels in this direction.")

        missing_sides_in_quarter = missing_sides | {"right", "back"}
        quarter_mesh = mesh_parallelepiped(
                size=(width/2, thickness/2, height), resolution=(nw//2, nth//2, nh),
                center=(-width/4, -thickness/4, 0), missing_sides=missing_sides_in_quarter,
                reflection_symmetry=False, translation_symmetry=False,
                name=f"quarter_of_{name}"
                )

        half_mesh = ReflectionSymmetricMesh(quarter_mesh, plane=yOz_Plane, name=f"half_of_{name}")
        mesh = ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=f"{name}")

    elif translation_symmetry:

        missing_sides_in_strip = missing_sides | {"left", "right"}
        strip = mesh_parallelepiped(
                size=(width/nw, thickness, height), resolution=(1, nth, nh),
                center=(-width/2 + width/(2*nw), 0, 0), missing_sides=missing_sides_in_strip,
                reflection_symmetry=False, translation_symmetry=False,
                name=f"strip_of_{name}"
                )

        open_parallelepiped = TranslationalSymmetricMesh(
            strip,
            translation=(width/nw, 0, 0), nb_repetitions=int(nw)-1,
            name=f"body_of_{name}"
        )

        components_of_mesh = [open_parallelepiped]
        if "right" not in missing_sides:
            components_of_mesh.append(
                    mesh_rectangle(
                        size=(thickness, height), resolution=(nth, nh),
                        center=(width/2, 0, 0), normal=(1, 0, 0),
                        name=f"right_side_of_{name}"
                        ))
        if "left" not in missing_sides:
            components_of_mesh.append(
                    mesh_rectangle(
                        size=(thickness, height), resolution=(nth, nh),
                        center=(-width/2, 0, 0), normal=(-1, 0, 0),
                        name=f"left_side_of_{name}"
                        ))

        mesh = CollectionOfMeshes(components_of_mesh, name=name)

    else:

        sides = []
        if "left" not in missing_sides:
            sides.append(
                    mesh_rectangle(size=(thickness, height), resolution=(nth, nh), center=(-width/2, 0, 0),
                        normal=(-1, 0, 0), name=f"left_of_{name}")
                    )
        if "right" not in missing_sides:
            sides.append(
                    mesh_rectangle(size=(thickness, height), resolution=(nth, nh), center=(width/2, 0, 0),
                        normal=(1, 0, 0), name=f"right_of_{name}")
                    )
        if "front" not in missing_sides:
            sides.append(
                    mesh_rectangle(size=(width, height), resolution=(nw, nh), center=(0, -thickness/2, 0),
                        normal=(0, -1, 0), name=f"front_of_{name}")
                    )
        if "back" not in missing_sides:
            sides.append(
                    mesh_rectangle(size=(width, height), resolution=(nw, nh), center=(0, thickness/2, 0),
                        normal=(0, 1, 0), name=f"back_of_{name}")
                    )
        if "top" not in missing_sides:
            sides.append(
                    mesh_rectangle(size=(thickness, width), resolution=(nth, nw), center=(0, 0, height/2),
                        normal=(0, 0, 1), name=f"top_of_{name}")
                    )
        if "bottom" not in missing_sides:
            sides.append(
                    mesh_rectangle(size=(thickness, width), resolution=(nth, nw), center=(0, 0, -height/2),
                        normal=(0, 0, -1), name=f"bottom_of_{name}")
                    )
        mesh = CollectionOfMeshes(sides, name=name).merged()

    mesh.heal_mesh()
    mesh.translate(center)
    mesh.geometric_center = np.asarray(center, dtype=float)
    return mesh



"""Generate meshes of cylinders and disks"""
# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging
from itertools import product

import numpy as np
from numpy import pi, cos, sin

from capytaine.meshes.geometry import xOz_Plane, yOz_Plane, Oz_axis
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import TranslationalSymmetricMesh, AxialSymmetricMesh, ReflectionSymmetricMesh

LOG = logging.getLogger(__name__)


def mesh_disk(*, radius=1.0, center=(0, 0, 0), normal=(0, 0, 1),
        resolution=(3, 6), faces_max_radius=None,
        reflection_symmetry=False, axial_symmetry=False, name=None, _theta_max=2*pi):
    """(One-sided) disk.

    Parameters
    ----------
    radius : float, optional
        radius of the disk
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the disk
    normal: 3-ple of floats, optional
        normal vector, default: along x axis
    resolution : 2-ple of int, optional
        number of panels along a radius and around the disk
    faces_max_radius : float, optional
        maximal radius of a panel. (Default: no maximal radius.)
        If the provided resolution is too coarse, the number of panels is
        changed to fit the constraint on the maximal radius.
    axial_symmetry : bool, optional
        if True, returns an AxialSymmetricMesh
    reflection_symmetry : bool, optional
        if True, returns a ReflectionSymmetricMesh
    name : str, optional
        a string naming the mesh
    _theta_max: float, optional
        internal parameter, to return an arc circle instead of a full circle
    """
    assert radius > 0, "Radius of the disk mesh should be given as a positive value."

    assert len(resolution) == 2, "Resolution of a disk should be given as a couple of values."
    assert all([h > 0 for h in resolution]), "Resolution of the disk mesh should be given as positive values."
    assert all([i == int(i) for i in resolution]), "Resolution of a disk should be given as integer values."

    assert len(center) == 3, "Position of the center of a disk should be given a 3-ple of values."

    nr, ntheta = resolution
    if faces_max_radius is not None:
        # The biggest cell is on the side of the disk.
        # Assuming it is a rectangle of sides radius/nr and perimeter/ntheta
        estimated_max_radius = np.hypot(radius/nr, 2*pi*radius/ntheta)/2
        if estimated_max_radius > faces_max_radius:
            nr = int(np.ceil(radius / (np.sqrt(2) * faces_max_radius)))
            ntheta = int(np.ceil(2*pi*radius / (np.sqrt(2) * faces_max_radius)))

    if name is None:
        name = f"disk_{next(Mesh._ids)}"

    if reflection_symmetry and axial_symmetry:
        raise NotImplementedError("Disks with both symmetries have not been implemented.")

    LOG.debug(f"New disk of radius {radius} and resolution {resolution}, named {name}.")

    if reflection_symmetry:
        if ntheta % 2 == 1:
            raise ValueError("To use the reflection symmetry of the mesh, "
                             "it should have an even number of panels in this direction.")

        half_mesh = mesh_disk(radius=radius, _theta_max=_theta_max/2, resolution=(nr, ntheta//2),
                center=(0, 0, 0), normal=(0, 0, 1),
                reflection_symmetry=False, axial_symmetry=False, name=f"half_of_{name}")
        mesh = ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=name)

    elif axial_symmetry:
        mesh_slice = mesh_disk(radius=radius, _theta_max=_theta_max/ntheta, resolution=(nr, 1),
                center=(0, 0, 0), normal=(0, 0, 1),
                reflection_symmetry=False, axial_symmetry=False, name=f"slice_of_{name}")
        mesh = AxialSymmetricMesh(mesh_slice, axis=Oz_axis, nb_repetitions=ntheta - 1, name=name)

    else:
        theta_range = np.linspace(0, _theta_max, ntheta+1)
        r_range = np.linspace(0.0, radius, nr+1)
        nodes = np.array([(0, r*sin(t), r*cos(t)) for (r, t) in product(r_range, theta_range)])
        panels = np.array([(j+i*(ntheta+1), j+1+i*(ntheta+1), j+1+(i+1)*(ntheta+1), j+(i+1)*(ntheta+1))
                                for (i, j) in product(range(0, nr), range(0, ntheta))])

        mesh = Mesh(nodes, panels, name=name)

    mesh.heal_mesh()
    mesh.translate(center)
    mesh.rotate_around_center_to_align_vectors(center, mesh.faces_normals[0], normal)
    mesh.geometric_center = np.asarray(center, dtype=float)
    return mesh


def mesh_vertical_cylinder(*, length=10.0, radius=1.0, center=(0, 0, 0),
        resolution=(2, 8, 10), faces_max_radius=None,
        axial_symmetry=False, reflection_symmetry=False, name=None, _theta_max=2*pi):
    """Vertical cylinder.

    Total number of panels = (2*resolution[0] + resolution[2])*resolution[1]

    Parameters
    ----------
    length : float, optional
        length of the cylinder
    radius : float, optional
        radius of the cylinder
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the cylinder
    resolution : 3-ple of int, optional
        (number of panel along a radius at the end, number of panels around a slice, number of slices)
        Mnemonic: same ordering as the cylindrical coordinates (nr, ntheta, nz)
    faces_max_radius : float, optional
        maximal radius of a panel. (Default: no maximal radius.)
        If the provided resolution is too coarse, the number of panels is
        changed to fit the constraint on the maximal radius.
    axial_symmetry : bool, optional
        if True, returns an AxialSymmetricMesh
    reflection_symmetry : bool, optional
        if True, returns a ReflectionSymmetricMesh
    name : str, optional
        a string naming the mesh
    _theta_max: float, optional
        internal parameter, to return an arc circle instead of a full circle
    """
    assert length > 0, "Length of a cylinder should be given as a positive value."
    assert radius > 0, "Radius of a cylinder should be given as a positive value."

    assert len(resolution) == 3, "Resolution of a cylinder should be given as a 3-ple of values."
    assert all([h >= 0 for h in resolution]), "Resolution of a cylinder should be given as positive values."
    assert all([i == int(i) for i in resolution]), "Resolution of a cylinder should be given as integer values."

    assert len(center) == 3, "Position of the center of a cylinder should be given a 3-ple of values."

    nr, ntheta, nz = resolution
    if faces_max_radius is not None:
        dr, dtheta, dz = radius/nr, 2*pi*radius/ntheta, length/nz
        estimated_max_radius = max(
                np.hypot(dr, dtheta)/2,  # Panel on disk side
                np.hypot(dtheta, dz)/2,  # Panel on cylinder itself
                )
        if estimated_max_radius > faces_max_radius:
            nr = int(np.ceil(radius / (np.sqrt(2) * faces_max_radius)))
            ntheta = 2*int(np.ceil(pi*radius / (np.sqrt(2) * faces_max_radius)))
            nz = int(np.ceil(length / (np.sqrt(2) * faces_max_radius)))

    if name is None:
        name = f"cylinder_{next(Mesh._ids)}"

    LOG.debug(f"New vertical cylinder of length {length}, radius {radius} and resolution {resolution}, named {name}.")

    if reflection_symmetry and axial_symmetry:
        raise NotImplementedError("Vertical cylinders with both symmetries have not been implemented.")

    if reflection_symmetry:
        if ntheta % 2 == 1:
            raise ValueError("To use the reflection symmetry of the mesh, "
                             "it should have an even number of panels in this direction.")

        half_cylinder = mesh_vertical_cylinder(length=length, radius=radius, center=(0, 0, 0),
                resolution=(nr, ntheta//2, nz), reflection_symmetry=False, axial_symmetry=False,
                name=f"half_{name}", _theta_max=_theta_max/2)

        mesh = ReflectionSymmetricMesh(half_cylinder, plane=xOz_Plane, name=name)

    elif axial_symmetry:

        mesh_slice = mesh_vertical_cylinder(length=length, radius=radius, resolution=(nr, 1, nz), center=(0, 0, 0),
                reflection_symmetry=False, axial_symmetry=False, name=f"slice_of_{name}", _theta_max=_theta_max/ntheta)
        mesh = AxialSymmetricMesh(mesh_slice, axis=Oz_axis, nb_repetitions=ntheta - 1, name=name)

    else:
        theta_range = np.linspace(0, _theta_max, ntheta+1)
        z_range = np.linspace(-length/2, length/2, nz+1)
        if nr > 0:
            r_range = np.linspace(0.0, radius, nr+1)
            nodes = np.concatenate([
                np.array([(r*sin(t), r*cos(t), -length/2) for (r, t) in product(r_range, theta_range)]),
                np.array([(radius*sin(t), radius*cos(t), z) for (z, t) in product(z_range, theta_range)]),
                np.array([(r*sin(t), r*cos(t), length/2) for (r, t) in product(r_range[::-1], theta_range)]),
                ])
        else:
            r_range = np.array([])
            nodes = np.array([(radius*sin(t), radius*cos(t), z) for (z, t) in product(z_range, theta_range)])
        panels = np.array([(j+i*(ntheta+1), j+(i+1)*(ntheta+1), j+1+(i+1)*(ntheta+1), j+1+i*(ntheta+1), )
                                for (i, j) in product(range(nz+2*(len(r_range))), range(ntheta))])

        mesh = Mesh(nodes, panels, name=name)

    mesh.heal_mesh()
    mesh.translate(center)
    mesh.geometric_center = np.asarray(center, dtype=float)
    return mesh


def mesh_horizontal_cylinder(*, length=10.0, radius=1.0, center=(0, 0, 0),
        resolution=(2, 8, 10), faces_max_radius=None,
        reflection_symmetry=False, translation_symmetry=False, name=None, _theta_max=2*pi):
    """Cylinder aligned along Ox axis.

    Total number of panels = (2*resolution[0] + resolution[2])*resolution[1]

    Parameters
    ----------
    length : float, optional
        length of the cylinder
    radius : float, optional
        radius of the cylinder
    center : 3-ple or array of shape (3,), optional
        position of the geometric center of the cylinder
    resolution : 3-ple of int, optional
        (number of panel along a radius at the end, number of panels around a slice, number of slices)
        Mnemonic: same ordering as the cylindrical coordinates (nr, ntheta, nz)
    faces_max_radius : float, optional
        maximal radius of a panel. (Default: no maximal radius.)
        If the provided resolution is too coarse, the number of panels is
        changed to fit the constraint on the maximal radius.
    reflection_symmetry : bool, optional
        if True, returns a ReflectionSymmetricMesh
    translation_symmetry : bool, optional
        if True, uses a TranslationalSymmetricMesh internally for the main part of the cylinder
    name : str, optional
        a string naming the mesh
    _theta_max: float, optional
        internal parameter, to return an arc circle instead of a full circle
    """

    assert length > 0, "Length of a cylinder should be given as a positive value."
    assert radius > 0, "Radius of a cylinder should be given as a positive value."

    assert len(resolution) == 3, "Resolution of a cylinder should be given as a 3-ple of values."
    assert all([h >= 0 for h in resolution]), "Resolution of a cylinder should be given as positive values."
    assert all([i == int(i) for i in resolution]), "Resolution of a cylinder should be given as integer values."

    assert len(center) == 3, "Position of the center of a cylinder should be given a 3-ple of values."

    if name is None:
        name = f"cylinder_{next(Mesh._ids)}"

    LOG.debug(f"New horizontal cylinder of length {length}, radius {radius} and resolution {resolution}, named {name}.")

    nr, ntheta, nx = resolution
    if faces_max_radius is not None:
        dr, dtheta, dx = radius/nr, 2*pi*radius/ntheta, length/nx
        estimated_max_radius = max(
                np.hypot(dr, dtheta)/2,  # Panel on disk side
                np.hypot(dtheta, dx)/2,  # Panel on cylinder itself
                )
        if estimated_max_radius > faces_max_radius:
            nr = int(np.ceil(radius / (np.sqrt(2) * faces_max_radius)))
            ntheta = 2*int(np.ceil(pi*radius / (np.sqrt(2) * faces_max_radius)))
            nx = int(np.ceil(length / (np.sqrt(2) * faces_max_radius)))

    if reflection_symmetry:
        if ntheta % 2 == 1:
            raise ValueError("To use the reflection symmetry of the mesh, "
                             "it should have an even number of panels in this direction.")

        half_cylinder = mesh_horizontal_cylinder(
                length=length, radius=radius, center=(0, 0, 0),
                resolution=(nr, ntheta//2, nx),
                reflection_symmetry=False, translation_symmetry=translation_symmetry,
                name=f"half_{name}", _theta_max=_theta_max/2,
                )

        mesh = ReflectionSymmetricMesh(half_cylinder, plane=xOz_Plane, name=name)

    else:
        if translation_symmetry:
            slice = mesh_horizontal_cylinder(
                    length=length/nx, radius=radius, center=(-length/2 + length/(2*nx), 0, 0),
                    resolution=(0, ntheta, 1),
                    reflection_symmetry=False, translation_symmetry=False,
                    name=f"slice_of_{name}", _theta_max=_theta_max,
                    )

            open_cylinder = TranslationalSymmetricMesh(
                    slice, translation=np.asarray([length / nx, 0.0, 0.0]),
                    nb_repetitions=nx-1, name=f"open_{name}")

        else: # General case
            theta_range = np.linspace(0, _theta_max, ntheta+1)
            x_range = np.linspace(-length/2, length/2, nx+1)
            nodes = np.array([(x, radius*sin(t), -radius*cos(t)) for (x, t) in product(x_range, theta_range)])

            panels = np.array([(i+j*(ntheta+1), i+1+j*(ntheta+1), i+1+(j+1)*(ntheta+1), i+(j+1)*(ntheta+1))
                                for (i, j) in product(range(ntheta), range(nx))])

            open_cylinder = Mesh(nodes, panels, name=f"open_{name}")

        if nr > 0:
            side = mesh_disk(radius=radius, center=(-length/2, 0, 0), normal=(-1, 0, 0),
                             reflection_symmetry=False, resolution=(nr, ntheta), name=f"side_of_{name}",
                             _theta_max=_theta_max)
            other_side = side.mirrored(yOz_Plane, name=f"other_side_of_{name}")
            mesh = CollectionOfMeshes([open_cylinder, side, other_side], name=name)
            if not translation_symmetry:
                mesh = mesh.merged()
        else:
            mesh = open_cylinder.copy(name=name)

    mesh.heal_mesh()
    mesh.translate(center)
    mesh.geometric_center = np.asarray(center, dtype=float)
    return mesh



import pytest
import numpy as np
import xarray as xr
import capytaine as cpt

try:
    import quadpy
    quadpy_methods = [quadpy.c2.schemes['sommariva_05']()]
except ImportError:
    quadpy_methods = []

from capytaine.meshes.quadratures import builtin_methods

@pytest.mark.parametrize('method', list(builtin_methods) + quadpy_methods)
def test_quadrature_points_are_coplanar(method):
    # Check that all quadrature points are within the plane containing the panel
    mesh = cpt.mesh_sphere()
    mesh.compute_quadrature(method)
    for i_face in range(mesh.nb_faces):
        A, B, C, D = mesh.vertices[mesh.faces[i_face, :], :]
        for j_quad_point in range(mesh.quadrature_points[0].shape[1]):
            X = mesh.quadrature_points[0][i_face, j_quad_point, :]
            assert np.isclose(np.dot(np.cross(B-A, C-A), X-A), 0.0, atol=1e-8)


@pytest.mark.parametrize('method', list(builtin_methods) + quadpy_methods)
def test_area(method):
    # Check that quadrature weights sum to the area of the panel
    mesh = cpt.mesh_sphere()
    mesh.compute_quadrature(method)
    for i_face in range(mesh.nb_faces):
        assert np.isclose(np.sum(mesh.quadrature_points[1][i_face, :]), mesh.faces_areas[i_face], rtol=1e-2)


def test_mesh_in_body():
    mesh = cpt.mesh_sphere()
    mesh.compute_quadrature("Gauss-Legendre 2")
    body = cpt.FloatingBody(mesh, cpt.rigid_body_dofs())
    assert body.mesh.quadrature_method == "Gauss-Legendre 2"
    points, weights = body.mesh.quadrature_points
    assert points.shape == (mesh.nb_faces, 4, 3)
    assert weights.shape == (mesh.nb_faces, 4)


@pytest.mark.parametrize('method', list(builtin_methods) + quadpy_methods)
def test_resolution(method):
    mesh = cpt.mesh_horizontal_cylinder(
        length=5.0, radius=1.0,
        center=(0, 0, 0),
        resolution=(2, 5, 10),
    ).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(name="Heave")

    test_matrix = xr.Dataset(coords={
        "omega": np.linspace(0.5, 3.0, 2),
        "radiating_dof": ["Heave"],
    })

    solver = cpt.BEMSolver()

    data_0 = solver.fill_dataset(test_matrix, [body], mesh=True)

    body.mesh.compute_quadrature(method)
    data_1 = solver.fill_dataset(test_matrix, [body], mesh=True)

    assert np.allclose(data_0["added_mass"].data, data_1["added_mass"].data, rtol=1e-1)


# def test_triangular_mesh()
#     mesh = cpt.Mesh(
#         vertices=np.array([[0.0, 0.0, -1.0], [1.0, 0.0, -1.0], [0.0, 1.0, -1.0], [0.0, 0.0, -1.0]]),
#         faces=np.array([[0, 1, 2, 3]])
#     )
#     mesh.compute_quadrature("GL2")
#     points, weights = mesh.quadrature_points



#!/usr/bin/env python

import capytaine as cpt
from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
from capytaine.ui.vtk.animation import Animation

cpt.set_logging('INFO')

# Generate the mesh of a sphere

full_mesh = cpt.mesh_sphere(radius=3, center=(0, 0, 0), resolution=(20, 20))
full_sphere = cpt.FloatingBody(mesh=full_mesh)
full_sphere.add_translation_dof(name="Heave")

# Keep only the immersed part of the mesh
sphere = full_sphere.immersed_part()

# Set up and solve problem
solver = cpt.BEMSolver()

diffraction_problem = cpt.DiffractionProblem(body=sphere, wave_direction=0.0, omega=2.0)
diffraction_result = solver.solve(diffraction_problem)

radiation_problem = cpt.RadiationProblem(body=sphere, radiating_dof="Heave", omega=2.0)
radiation_result = solver.solve(radiation_problem)

# Define a mesh of the free surface and compute the free surface elevation
free_surface = cpt.FreeSurface(x_range=(-50, 50), y_range=(-50, 50), nx=150, ny=150)
diffraction_elevation_at_faces = solver.compute_free_surface_elevation(free_surface, diffraction_result)
radiation_elevation_at_faces = solver.compute_free_surface_elevation(free_surface, radiation_result)

# Add incoming waves
diffraction_elevation_at_faces = diffraction_elevation_at_faces + airy_waves_free_surface_elevation(free_surface, diffraction_problem)

# Run the animations
animation = Animation(loop_duration=diffraction_result.period)
animation.add_body(full_sphere, faces_motion=None)
animation.add_free_surface(free_surface, faces_elevation=0.5*diffraction_elevation_at_faces)
animation.run(camera_position=(-30, -30, 30))  # The camera is oriented towards (0, 0, 0) by default.
# animation.save("path/to/the/video/file.ogv", camera_position=(-30, -30, 30))

animation = Animation(loop_duration=radiation_result.period)
animation.add_body(full_sphere, faces_motion=full_sphere.dofs["Heave"])
animation.add_free_surface(free_surface, faces_elevation=3.0*radiation_elevation_at_faces)
animation.run(camera_position=(-30, -30, 30))
# animation.save("path/to/the/video/file.ogv", camera_position=(-30, -30, 30))



"""Definition of the methods to build influence matrices, using possibly some sparse structures."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from abc import ABC, abstractmethod

import numpy as np
from scipy.linalg import lu_factor
from scipy.sparse import coo_matrix
from scipy.sparse import linalg as ssl

from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import ReflectionSymmetricMesh, TranslationalSymmetricMesh, AxialSymmetricMesh

from capytaine.matrices import linear_solvers
from capytaine.matrices.block import BlockMatrix
from capytaine.matrices.low_rank import LowRankMatrix, NoConvergenceOfACA
from capytaine.matrices.block_toeplitz import BlockSymmetricToeplitzMatrix, BlockToeplitzMatrix, BlockCirculantMatrix
from capytaine.tools.lru_cache import lru_cache_with_strict_maxsize

LOG = logging.getLogger(__name__)


####################
#  ABSTRACT CLASS  #
####################

class MatrixEngine(ABC):
    """Abstract method to build a matrix."""

    @abstractmethod
    def build_matrices(self, mesh1, mesh2, free_surface, water_depth, wavenumber, green_function, adjoint_double_layer):
        pass

    def build_S_matrix(self, *args, **kwargs):
        """Similar to :code:`build_matrices`, but returning only :math:`S`"""
        S, _ = self.build_matrices(*args, **kwargs)  # Could be optimized...
        return S


##################
#  BASIC ENGINE  #
##################

class BasicMatrixEngine(MatrixEngine):
    """
    Simple engine that assemble a full matrix (except for one reflection symmetry).
    Basically only calls :code:`green_function.evaluate`.

    Parameters
    ----------
    linear_solver: str or function, optional
        Setting of the numerical solver for linear problems Ax = b.
        It can be set with the name of a preexisting solver
        (available: "direct" and "gmres", the former is the default choice)
        or by passing directly a solver function.
    matrix_cache_size: int, optional
        number of matrices to keep in cache
    """

    available_linear_solvers = {'direct': linear_solvers.solve_directly,
                                'lu_decomposition': linear_solvers.LUSolverWithCache().solve,
                                'gmres': linear_solvers.solve_gmres,
                                }

    def __init__(self, *, linear_solver='lu_decomposition', matrix_cache_size=1):

        if linear_solver in self.available_linear_solvers:
            self.linear_solver = self.available_linear_solvers[linear_solver]
        else:
            self.linear_solver = linear_solver

        if matrix_cache_size > 0:
            self.build_matrices = lru_cache_with_strict_maxsize(maxsize=matrix_cache_size)(self.build_matrices)

        self.exportable_settings = {
            'engine': 'BasicMatrixEngine',
            'matrix_cache_size': matrix_cache_size,
            'linear_solver': str(linear_solver),
        }

    def __str__(self):
        params = f"linear_solver=\'{self.exportable_settings['linear_solver']}\'"
        params += f", matrix_cache_size={self.exportable_settings['matrix_cache_size']}" if self.exportable_settings['matrix_cache_size'] != 1 else ""
        return f"BasicMatrixEngine({params})"

    def __repr__(self):
        return self.__str__()

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    def build_matrices(self, mesh1, mesh2, free_surface, water_depth, wavenumber, green_function, adjoint_double_layer=True):
        r"""Build the influence matrices between mesh1 and mesh2.

        Parameters
        ----------
        mesh1: Mesh or CollectionOfMeshes
            mesh of the receiving body (where the potential is measured)
        mesh2: Mesh or CollectionOfMeshes
            mesh of the source body (over which the source distribution is integrated)
        free_surface: float
            position of the free surface (default: :math:`z = 0`)
        water_depth: float
            position of the sea bottom (default: :math:`z = -\infty`)
        wavenumber: float
            wavenumber (default: 1.0)
        green_function: AbstractGreenFunction
            object with an "evaluate" method that computes the Green function.
        adjoint_double_layer: bool, optional
            compute double layer for direct method (F) or adjoint double layer for indirect method (T) matrices (default: True)

        Returns
        -------
        tuple of matrix-like
            the matrices :math:`S` and :math:`K`
        """

        if (isinstance(mesh1, ReflectionSymmetricMesh)
                and isinstance(mesh2, ReflectionSymmetricMesh)
                and mesh1.plane == mesh2.plane):

            S_a, V_a = self.build_matrices(
                mesh1[0], mesh2[0], free_surface, water_depth, wavenumber,
                green_function)
            S_b, V_b = self.build_matrices(
                mesh1[0], mesh2[1], free_surface, water_depth, wavenumber,
                green_function)

            return BlockSymmetricToeplitzMatrix([[S_a, S_b]]), BlockSymmetricToeplitzMatrix([[V_a, V_b]])

        else:
            return green_function.evaluate(
                mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer=adjoint_double_layer
            )

###################################
#  HIERARCHIAL TOEPLITZ MATRICES  #
###################################

class HierarchicalToeplitzMatrixEngine(MatrixEngine):
    """An experimental matrix engine that build a hierarchical matrix with
     some block-Toeplitz structure.

    Parameters
    ----------
    ACA_distance: float, optional
        Above this distance, the ACA is used to approximate the matrix with a low-rank block.
    ACA_tol: float, optional
        The tolerance of the ACA when building a low-rank matrix.
    matrix_cache_size: int, optional
        number of matrices to keep in cache
    """

    def __init__(self, *, ACA_distance=8.0, ACA_tol=1e-2, matrix_cache_size=1):

        if matrix_cache_size > 0:
            self.build_matrices = lru_cache_with_strict_maxsize(maxsize=matrix_cache_size)(self.build_matrices)

        self.ACA_distance = ACA_distance
        self.ACA_tol = ACA_tol

        self.linear_solver = linear_solvers.solve_gmres

        self.exportable_settings = {
            'engine': 'HierarchicalToeplitzMatrixEngine',
            'ACA_distance': ACA_distance,
            'ACA_tol': ACA_tol,
            'matrix_cache_size': matrix_cache_size,
        }

    def __str__(self):
        params = f"ACA_distance={self.ACA_distance}"
        params += f", ACA_tol={self.ACA_tol}"
        params += f", matrix_cache_size={self.exportable_settings['matrix_cache_size']}" if self.exportable_settings['matrix_cache_size'] != 1 else ""
        return f"HierarchicalToeplitzMatrixEngine({params})"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())


    def build_matrices(self,
                       mesh1, mesh2, free_surface, water_depth, wavenumber, green_function,
                       adjoint_double_layer=True):

        return self._build_matrices(
                           mesh1, mesh2, free_surface, water_depth, wavenumber, green_function,
                           adjoint_double_layer, _rec_depth=1)


    def _build_matrices(self,
                       mesh1, mesh2, free_surface, water_depth, wavenumber, green_function,
                       adjoint_double_layer, _rec_depth=1):
        """Recursively builds a hierarchical matrix between mesh1 and mesh2.

        Same arguments as :func:`BasicMatrixEngine.build_matrices`.

        :code:`_rec_depth` keeps track of the recursion depth only for pretty log printing.
        """

        if logging.getLogger().isEnabledFor(logging.DEBUG):
            log_entry = (
                "\t" * (_rec_depth+1) +
                "Build the S and K influence matrices between {mesh1} and {mesh2}"
                .format(mesh1=mesh1.name, mesh2=(mesh2.name if mesh2 is not mesh1 else 'itself'))
            )
        else:
            log_entry = ""  # will not be used

        # Distance between the meshes (for ACA).
        distance = np.linalg.norm(mesh1.center_of_mass_of_nodes - mesh2.center_of_mass_of_nodes)

        # I) SPARSE COMPUTATION
        # I-i) BLOCK TOEPLITZ MATRIX

        if (isinstance(mesh1, ReflectionSymmetricMesh)
                and isinstance(mesh2, ReflectionSymmetricMesh)
                and mesh1.plane == mesh2.plane):

            LOG.debug(log_entry + " using mirror symmetry.")

            S_a, V_a = self._build_matrices(
                mesh1[0], mesh2[0], free_surface, water_depth, wavenumber, green_function,
                adjoint_double_layer=adjoint_double_layer, _rec_depth=_rec_depth+1)
            S_b, V_b = self._build_matrices(
                mesh1[0], mesh2[1], free_surface, water_depth, wavenumber, green_function,
                adjoint_double_layer=adjoint_double_layer, _rec_depth=_rec_depth+1)

            return BlockSymmetricToeplitzMatrix([[S_a, S_b]]), BlockSymmetricToeplitzMatrix([[V_a, V_b]])

        elif (isinstance(mesh1, TranslationalSymmetricMesh)
              and isinstance(mesh2, TranslationalSymmetricMesh)
              and np.allclose(mesh1.translation, mesh2.translation)
              and mesh1.nb_submeshes == mesh2.nb_submeshes):

            LOG.debug(log_entry + " using translational symmetry.")

            S_list, V_list = [], []
            for submesh in mesh2:
                S, V = self._build_matrices(
                    mesh1[0], submesh, free_surface, water_depth, wavenumber, green_function,
                    adjoint_double_layer=adjoint_double_layer, _rec_depth=_rec_depth+1)
                S_list.append(S)
                V_list.append(V)
            for submesh in mesh1[1:][::-1]:
                S, V = self._build_matrices(
                    submesh, mesh2[0], free_surface, water_depth, wavenumber, green_function,
                    adjoint_double_layer=adjoint_double_layer, _rec_depth=_rec_depth+1)
                S_list.append(S)
                V_list.append(V)

            return BlockToeplitzMatrix([S_list]), BlockToeplitzMatrix([V_list])

        elif (isinstance(mesh1, AxialSymmetricMesh)
              and isinstance(mesh2, AxialSymmetricMesh)
              and mesh1.axis == mesh2.axis
              and mesh1.nb_submeshes == mesh2.nb_submeshes):

            LOG.debug(log_entry + " using rotation symmetry.")

            S_line, V_line = [], []
            for submesh in mesh2[:mesh2.nb_submeshes]:
                S, V = self._build_matrices(
                    mesh1[0], submesh, free_surface, water_depth, wavenumber, green_function,
                    adjoint_double_layer=adjoint_double_layer, _rec_depth=_rec_depth+1)
                S_line.append(S)
                V_line.append(V)

            return BlockCirculantMatrix([S_line]), BlockCirculantMatrix([V_line])

        # I-ii) LOW-RANK MATRIX WITH ACA

        elif distance > self.ACA_distance*mesh1.diameter_of_nodes or distance > self.ACA_distance*mesh2.diameter_of_nodes:

            LOG.debug(log_entry + " using ACA.")

            def get_row_func(i):
                s, v = green_function.evaluate(
                    mesh1.extract_one_face(i), mesh2,
                    free_surface, water_depth, wavenumber,
                    adjoint_double_layer=adjoint_double_layer
                )
                return s.flatten(), v.flatten()

            def get_col_func(j):
                s, v = green_function.evaluate(
                    mesh1, mesh2.extract_one_face(j),
                    free_surface, water_depth, wavenumber,
                    adjoint_double_layer=adjoint_double_layer
                )
                return s.flatten(), v.flatten()

            try:
                return LowRankMatrix.from_rows_and_cols_functions_with_multi_ACA(
                    get_row_func, get_col_func, mesh1.nb_faces, mesh2.nb_faces,
                    nb_matrices=2, id_main=1,  # Approximate V and get an approximation of S at the same time
                    tol=self.ACA_tol, dtype=np.complex128)
            except NoConvergenceOfACA:
                pass  # Continue with non sparse computation

        # II) NON-SPARSE COMPUTATIONS
        # II-i) BLOCK MATRIX

        if (isinstance(mesh1, CollectionOfMeshes)
              and isinstance(mesh2, CollectionOfMeshes)):

            LOG.debug(log_entry + " using block matrix structure.")

            S_matrix, V_matrix = [], []
            for submesh1 in mesh1:
                S_line, V_line = [], []
                for submesh2 in mesh2:
                    S, V = self._build_matrices(
                        submesh1, submesh2, free_surface, water_depth, wavenumber, green_function,
                        adjoint_double_layer=adjoint_double_layer, _rec_depth=_rec_depth+1)

                    S_line.append(S)
                    V_line.append(V)
                S_matrix.append(S_line)
                V_matrix.append(V_line)

            return BlockMatrix(S_matrix), BlockMatrix(V_matrix)

        # II-ii) PLAIN NUMPY ARRAY

        else:
            LOG.debug(log_entry)

            S, V = green_function.evaluate(
                mesh1, mesh2, free_surface, water_depth, wavenumber, adjoint_double_layer=adjoint_double_layer
            )
            return S, V

class HierarchicalPrecondMatrixEngine(HierarchicalToeplitzMatrixEngine):
    """An experimental matrix engine that build a hierarchical matrix with
     some block-Toeplitz structure.

    Parameters
    ----------
    ACA_distance: float, optional
        Above this distance, the ACA is used to approximate the matrix with a low-rank block.
    ACA_tol: float, optional
        The tolerance of the ACA when building a low-rank matrix.
    matrix_cache_size: int, optional
        number of matrices to keep in cache
    """

    def __init__(self, *, ACA_distance=8.0, ACA_tol=1e-2, matrix_cache_size=1):
        super().__init__(ACA_distance=ACA_distance, ACA_tol=ACA_tol, matrix_cache_size=matrix_cache_size)
        self.linear_solver = linear_solvers.solve_precond_gmres

    def build_matrices(self,
                       mesh1, mesh2, free_surface, water_depth, wavenumber,
                       green_function, adjoint_double_layer=True):
        """Recursively builds a hierarchical matrix between mesh1 and mesh2,
        and precomputes some of the quantities needed for the preconditioner.

        Same arguments as :func:`BasicMatrixEngine.build_matrices`, except for rec_depth
        """
        # Build the matrices using the method of the parent class
        S, K = super().build_matrices(mesh1, mesh2, free_surface, water_depth,
                                      wavenumber, green_function,
                                      adjoint_double_layer=adjoint_double_layer)

        path_to_leaf = mesh1.path_to_leaf()

        n = len(path_to_leaf)
        N = K.shape[0]

        # Navigate to the diagonal blocks and compute their LU decompositions
        DLU = []
        diag_shapes = []
        for leaf in range(n):
            # Navigate to the block containing the one we need
            # (one layer above in the dendrogram)
            #upper_block = self.access_block_by_path(K, path_to_leaf[leaf][:-1])
            upper_block = K.access_block_by_path(path_to_leaf[leaf][:-1])
            # find the local index in the full path
            ind = path_to_leaf[leaf][-1]
            # compute the LU decomposition and add to the list
            DLU.append(lu_factor(upper_block.all_blocks[ind, ind]))
            diag_shapes.append(upper_block.all_blocks[ind, ind].shape[0])

        # Build the restriction and precompute its multiplication by K
        R = np.zeros((n, N), dtype=complex)
        RA = np.zeros((n, N), dtype=complex)
        for ii in range(n):
            row_slice = slice(sum(diag_shapes[:ii]), sum(diag_shapes[:ii+1]))
            R[ii, row_slice] = 1
            # Compute the multiplication using only the relevant slices of K
            # The slices are found by navigating the tree
            #RA[ii, :] = self.slice_rmatvec(R[ii, :], ii)
            Aloc = K
            v = R[ii, :]
            va = np.zeros(N, dtype=complex)
            free = [0, N]

            for lvl, jj in enumerate(path_to_leaf[ii]):

                Nrows = Aloc.all_blocks[jj, jj].shape[0]

                if jj==0:
                    v = v[:Nrows]
                    w = v @ Aloc.all_blocks[0,1]
                    va[free[1]-len(w) : free[1]] = w
                    free[1] = free[1] - len(w)
                else:
                    v = v[-Nrows:]
                    w = v @ Aloc.all_blocks[1, 0]
                    va[free[0] : free[0]+len(w)] = w
                    free[0] = free[0] + len(w)

                Aloc = Aloc.all_blocks[jj, jj]

                if lvl == len(path_to_leaf[ii])-1:
                    w = v@Aloc
                    va[free[0] : free[1]] = w
                    free[0] = free[0] + len(w)

            RA[ii, :] = va

        Ac = RA @ R.T
        AcLU = lu_factor(Ac)

        # Now navigate again to the diagonal blocks and set them to zero
        for leaf in range(n):
            upper_block = K.access_block_by_path(path_to_leaf[leaf][:-1])
            ind = path_to_leaf[leaf][-1]
            # turn the diagonal block into a zero sparse matrix
            upper_block.all_blocks[ind, ind] = coo_matrix(upper_block.all_blocks[ind, ind].shape)

        def PinvA_mv(v):
            v = v + 1j*np.zeros(N)
            return v - linear_solvers._block_Jacobi_coarse_corr(
                             K, np.zeros(N, dtype=complex), v,
                             R, RA, AcLU, DLU, diag_shapes, n)

        PinvA = ssl.LinearOperator((N, N), matvec=PinvA_mv)

        return S, (K, R, RA, AcLU, DLU, diag_shapes, n, PinvA)



import numpy as np
import capytaine as cpt
import pytest

method = ['indirect', 'direct']

@pytest.mark.parametrize("method", method)
def test_sum_of_dofs(method):
    body1 = cpt.Sphere(radius=1.0, ntheta=3, nphi=12, center=(0, 0, -3), name="body1")
    body1.add_translation_dof(name="Heave")

    body2 = cpt.Sphere(radius=1.0, ntheta=3, nphi=8,center=(5, 3, -1.5), name="body2")
    body2.add_translation_dof(name="Heave")

    both = body1 + body2
    both.add_translation_dof(name="Heave")

    problems = [cpt.RadiationProblem(body=both, radiating_dof=dof, omega=1.0) for dof in both.dofs]
    solver = cpt.BEMSolver()
    results = solver.solve_all(problems, method=method)
    dataset = cpt.assemble_dataset(results)

    both_added_mass = dataset['added_mass'].sel(radiating_dof="Heave", influenced_dof="Heave").data
    body1_added_mass = dataset['added_mass'].sel(radiating_dof="body1__Heave", influenced_dof="body1__Heave").data
    body2_added_mass = dataset['added_mass'].sel(radiating_dof="body2__Heave", influenced_dof="body2__Heave").data

    assert np.allclose(both_added_mass, body1_added_mass + body2_added_mass, rtol=1e-2)


@pytest.mark.parametrize("method", method)
def test_rotation_axis(method):
    body = cpt.RectangularParallelepiped(resolution=(4, 4, 4), center=(0, 0, -1), name="body")
    body.add_translation_dof(name="Sway")
    body.add_rotation_dof(axis=cpt.Axis(point=(0, 0, 0), vector=(0, 0, 1)), name="Yaw")

    l = 2.0
    body.add_rotation_dof(axis=cpt.Axis(point=(l, 0, 0), vector=(0, 0, 1)), name="other_rotation")

    assert np.allclose(body.dofs['other_rotation'], (body.dofs['Yaw'] - l*body.dofs['Sway']))

    problems = [cpt.RadiationProblem(body=body, radiating_dof=dof, omega=1.0) for dof in body.dofs]
    solver = cpt.BEMSolver()
    results = solver.solve_all(problems, keep_details=True, method=method)
    dataset = cpt.assemble_dataset(results)

    if ( method == 'indirect' ):
      sources = {result.radiating_dof: result.sources for result in results}
      assert np.allclose(sources['other_rotation'],
                       sources['Yaw'] - l*sources['Sway'], atol=1e-4)

    potential = {result.radiating_dof: result.potential for result in results}
    assert np.allclose(potential['other_rotation'],
                       potential['Yaw'] - l*potential['Sway'], atol=1e-4)

    A_m = dataset['added_mass'].sel(radiating_dof="other_rotation", influenced_dof="other_rotation").data
    A = dataset['added_mass'].sel(radiating_dof=["Yaw", "Sway"], influenced_dof=["Yaw", "Sway"]).data
    P = np.array([1, -l])
    assert np.isclose(A_m, P.T @ A @ P)



# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

from capytaine.meshes.predefined.cylinders import mesh_disk, mesh_horizontal_cylinder, mesh_vertical_cylinder
from capytaine.meshes.predefined.spheres import mesh_sphere
from capytaine.meshes.predefined.rectangles import mesh_rectangle, mesh_parallelepiped



# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import ReflectionSymmetricMesh, TranslationalSymmetricMesh, AxialSymmetricMesh



import pytest
import numpy as np
import capytaine as cpt
from capytaine.tools.symbolic_multiplication import SymbolicMultiplication, supporting_symbolic_multiplication

def test_definition():
    zero = SymbolicMultiplication("0")
    assert zero.symbol == "0"
    assert isinstance(zero, SymbolicMultiplication)

def test_multiplication():
    zero = SymbolicMultiplication("0")
    b = 2 * zero
    assert b.value == 2 * zero.value

def test_division():
    zero = SymbolicMultiplication("0")
    b = 2 * zero
    assert b / zero == 2

def test_double_division():
    zero = SymbolicMultiplication("0")
    b = 7 * zero * zero
    assert b / (zero * zero) == 7

def test_invert_division():
    zero = SymbolicMultiplication("0")
    b = 9 * zero
    assert zero / b == pytest.approx(1/9)

def test_float():
    zero = SymbolicMultiplication("0")
    b = 4.5 * zero
    assert float(b) == pytest.approx(0.0)

def test_numpy_array():
    zero = SymbolicMultiplication("0")
    b = np.random.rand(10) * zero
    assert (b / zero).shape == (10,)

def test_numpy_array_sum():
    zero = SymbolicMultiplication("0")
    b = np.ones(10) * zero
    assert (np.sum(b) / zero) == 10

def test_numpy_matmul():
    zero = SymbolicMultiplication("0")
    b = np.random.rand(10) * zero
    A = np.random.rand(10, 10)
    c = A @ b
    assert (c/zero).shape == (10,)

def test_supporting_symbolic_multiplication():
    zero = SymbolicMultiplication("0")

    @supporting_symbolic_multiplication
    def my_linear_operator(A, x):
        return np.linalg.solve(A, x)

    b = np.random.rand(10) * zero
    A = np.random.rand(10, 10)
    c = my_linear_operator(A, b)
    assert (c/zero).shape == (10,)

@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_radiation_problem_definition_for_zero_freq(water_depth):
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, omega=0.0, water_depth=water_depth, radiating_dof="Heave")
    assert float(pb.wavelength) == np.inf

@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_radiation_problem_definition_for_inf_freq(water_depth):
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, omega=np.inf, water_depth=water_depth, radiating_dof="Heave")
    assert float(pb.wavelength) == 0.0

@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_radiation_problem_definition_for_zero_wavelength(water_depth):
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, wavelength=0.0, water_depth=water_depth, radiating_dof="Heave")
    assert float(pb.period) == 0.0
    assert float(pb.wavenumber) == np.inf

@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_radiation_problem_definition_for_inf_wavelength(water_depth):
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, wavelength=np.inf, water_depth=water_depth, radiating_dof="Heave")
    assert float(pb.period) == np.inf
    assert float(pb.wavenumber) == 0.0

@pytest.mark.parametrize("omega", [0.0, np.inf])
def test_radiation_problem_definition(omega):
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, omega=omega, water_depth=np.inf, radiating_dof="Heave")
    res = cpt.BEMSolver().solve(pb)
    assert type(res.added_masses["Heave"]) == float

@pytest.mark.parametrize("omega", [0.0, np.inf])
def test_diffraction_problem(omega):
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    with pytest.raises(NotImplementedError):
        pb = cpt.DiffractionProblem(body=body, omega=omega)



#!/usr/bin/env python

import numpy as np
from numpy import dot, pi
from numpy.linalg import norm

import capytaine as cpt
from capytaine.green_functions.abstract_green_function import AbstractGreenFunction

class MyGreenFunction(AbstractGreenFunction):
    """An example of a custom routine to evaluate the Green function."""

    def evaluate(self, mesh1, mesh2, free_surface, water_depth, wavenumber, **kwargs):
        """The main method that needs to be implemented in the class."""

        if free_surface == np.inf and water_depth == np.inf:

            # Initialize the matrices
            S = np.zeros((mesh1.nb_faces, mesh2.nb_faces))
            K = np.zeros((mesh1.nb_faces, mesh2.nb_faces))

            for i in range(mesh1.nb_faces):
                for j in range(mesh2.nb_faces):
                    area = mesh1.faces_areas[i]
                    if i != j:
                        # The integral on the face is computed using the value at the center of the face.
                        r = mesh1.faces_centers[i, :] - mesh2.faces_centers[j, :]

                        S[j, i] = -area/(4*pi*norm(r))
                        K[j, i] = -area*dot(r, mesh1.faces_normals[j, :])/(4*pi*norm(r)**2)

                    else:
                        # The kernel is singular.
                        # When i == j, we can't use the approximation above for the integral.
                        # Below, we take a rough approximation.
                        S[j, i] = -area/(4*pi)*3
                        K[j, i] = -area/(4*pi)*3

            if mesh1 is mesh2:
                for i in range(mesh1.nb_faces):
                    K[i, i] += 1/2

            return S, K

        else:
            raise NotImplementedError

# Define a solver using the Green function above.
solver = cpt.BEMSolver(green_function=MyGreenFunction())

# Example of a problem
sphere = cpt.FloatingBody(
        mesh=cpt.mesh_sphere(
            radius=1.0,          # Dimension
            center=(0, 0, -2),   # Position
            resolution=(4, 4),   # Fineness of the mesh
            ))
sphere.add_translation_dof(name="Heave")
problem = cpt.RadiationProblem(body=sphere, free_surface=np.inf, water_depth=np.inf, radiating_dof='Heave')

result = solver.solve(problem)
print(result.added_masses)



"""Definition of the problems to solve with the BEM solver, and the results of this resolution."""
# Copyright (C) 2017-2023 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging

import numpy as np
import pandas as pd
from scipy.optimize import newton

from capytaine.tools.deprecation_handling import _get_water_depth
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.bem.airy_waves import airy_waves_velocity, froude_krylov_force
from capytaine.tools.symbolic_multiplication import SymbolicMultiplication

LOG = logging.getLogger(__name__)

_default_parameters = {'rho': 1000.0, 'g': 9.81, 'omega': 1.0,
                      'free_surface': 0.0, 'water_depth': np.inf,
                       'wave_direction': 0.0, 'forward_speed': 0.0}



class LinearPotentialFlowProblem:
    """General class of a potential flow problem.

    At most one of the following parameter must be provided: omega, period, wavenumber or wavelength.
    Internally only omega is stored, hence setting another parameter can lead to small rounding errors.

    Parameters
    ----------
    body: FloatingBody, optional
        The body interacting with the waves
    free_surface: float, optional
        The position of the free surface (accepted values: 0 and np.inf)
    water_depth: float, optional
        The depth of water in m (default: np.inf)
    sea_bottom: float, optional
        The position of the sea bottom (deprecated: please prefer setting water_depth)
    omega: float, optional
        The angular frequency of the waves in rad/s
    period: float, optional
        The period of the waves in s
    wavenumber: float, optional
        The angular wave number of the waves in rad/m
    wavelength: float, optional
        The wave length of the waves in m
    forward_speed: float, optional
        The speed of the body (in m/s, in the x direction, default: 0.0)
    rho: float, optional
        The density of water in kg/m3 (default: 1000.0)
    g: float, optional
        The acceleration of gravity in m/s2 (default: 9.81)
    boundary_condition: np.ndarray of shape (body.mesh.nb_faces,), optional
        The Neumann boundary condition on the floating body
    """

    def __init__(self, *,
                 body=None,
                 free_surface=_default_parameters['free_surface'],
                 water_depth=None, sea_bottom=None,
                 omega=None, period=None, wavenumber=None, wavelength=None,
                 forward_speed=_default_parameters['forward_speed'],
                 rho=_default_parameters['rho'],
                 g=_default_parameters['g'],
                 wave_direction=_default_parameters['wave_direction'],
                 boundary_condition=None):

        self.body = body
        self.free_surface = float(free_surface)
        self.rho = float(rho)
        self.g = float(g)
        self.forward_speed = float(forward_speed)
        self.wave_direction = float(wave_direction)  # Required for (diffraction problem) and (radiation problems with forward speed).

        self.boundary_condition = boundary_condition

        self.water_depth = _get_water_depth(free_surface, water_depth, sea_bottom, default_water_depth=_default_parameters["water_depth"])
        self.omega, self.period, self.wavenumber, self.wavelength, self.provided_freq_type = \
                self._get_frequencies(omega=omega, period=period, wavenumber=wavenumber, wavelength=wavelength)

        self._check_data()

        if forward_speed != 0.0:
            dopplered_omega = self.omega - self.wavenumber*self.forward_speed*np.cos(self.wave_direction)
            self.encounter_omega, self.encounter_period, self.encounter_wavenumber, self.encounter_wavelength, _ = \
                    self._get_frequencies(omega=abs(dopplered_omega))

            if dopplered_omega >= 0.0:
                self.encounter_wave_direction = self.wave_direction
            else:
                self.encounter_wave_direction = self.wave_direction + np.pi
        else:
            self.encounter_omega = self.omega
            self.encounter_period = self.period
            self.encounter_wavenumber = self.wavenumber
            self.encounter_wavelength = self.wavelength
            self.encounter_wave_direction = self.wave_direction


    def _get_frequencies(self, *, omega=None, period=None, wavenumber=None, wavelength=None):
        frequency_data = dict(omega=omega, period=period, wavenumber=wavenumber, wavelength=wavelength)
        nb_provided_frequency_data = 4 - list(frequency_data.values()).count(None)

        if nb_provided_frequency_data > 1:
            raise ValueError("Settings a problem requires at most one of the following: omega (angular frequency) OR period OR wavenumber OR wavelength.\n"
                             "Received {} of them: {}".format(nb_provided_frequency_data, {k: v for k, v in frequency_data.items() if v is not None}))

        if nb_provided_frequency_data == 0:
            provided_freq_type = 'omega'
            frequency_data = {'omega': _default_parameters['omega']}
        else:
            provided_freq_type = [k for (k, v) in frequency_data.items() if v is not None][0]

        if ((float(frequency_data[provided_freq_type]) == 0.0 and provided_freq_type in {'omega', 'wavenumber'})
            or (float(frequency_data[provided_freq_type]) == np.inf and provided_freq_type in {'period', 'wavelength'})):
                omega = SymbolicMultiplication("0")
                wavenumber = SymbolicMultiplication("0")
                period = SymbolicMultiplication("∞")
                wavelength = SymbolicMultiplication("∞")
        elif ((float(frequency_data[provided_freq_type]) == 0.0 and provided_freq_type in {'period', 'wavelength'})
            or (float(frequency_data[provided_freq_type]) == np.inf and provided_freq_type in {'omega', 'wavenumber'})):
                omega = SymbolicMultiplication("∞")
                wavenumber = SymbolicMultiplication("∞")
                period = SymbolicMultiplication("0")
                wavelength = SymbolicMultiplication("0")
        else:

            if provided_freq_type in {'omega', 'period'}:
                if provided_freq_type == 'omega':
                    omega = frequency_data['omega']
                    period = 2*np.pi/omega
                else:  # provided_freq_type is 'period'
                    period = frequency_data['period']
                    omega = 2*np.pi/period

                if self.water_depth == np.inf:
                    wavenumber = omega**2/self.g
                else:
                    wavenumber = newton(lambda k: k*np.tanh(k*self.water_depth) - omega**2/self.g, x0=1.0)
                wavelength = 2*np.pi/wavenumber

            else:  # provided_freq_type is 'wavelength' or 'wavenumber'
                if provided_freq_type == 'wavelength':
                    wavelength = frequency_data['wavelength']
                    wavenumber = 2*np.pi/wavelength
                else:  # provided_freq_type is 'wavenumber'
                    wavenumber = frequency_data['wavenumber']
                    wavelength = 2*np.pi/wavenumber

                omega = np.sqrt(self.g*wavenumber*np.tanh(wavenumber*self.water_depth))
                period = 2*np.pi/omega

        return omega, period, wavenumber, wavelength, provided_freq_type

    def _check_data(self):
        """Sanity checks on the data."""

        if self.free_surface not in {0.0, np.inf}:
            raise NotImplementedError(
                f"Free surface is {self.free_surface}. "
                "Only z=0 and z=∞ are accepted values for the free surface position."
            )

        if not (-2*np.pi-1e-3 <= self.wave_direction <= 2*np.pi+1e-3):
            LOG.warning(f"The value {self.wave_direction} has been provided for the wave direction, and it does not look like an angle in radians. "
                         "The wave direction in Capytaine is defined in radians and not in degrees, so the result might not be what you expect. "
                         "If you were actually giving an angle in radians, use the modulo operator to give a value between -2π and 2π to disable this warning.")

        if self.free_surface == np.inf and self.water_depth != np.inf:

            raise NotImplementedError(
                "Problems with a sea bottom but no free surface have not been implemented."
            )

        if self.water_depth < 0.0:
            raise ValueError("`water_depth` should be strictly positive (provided water depth: {self.water_depth}).")

        if float(self.omega) in {0, np.inf}:
            if self.water_depth != np.inf:
                LOG.warning(
                        f"Default Green function allows for {self.provided_freq_type}={float(self.__getattribute__(self.provided_freq_type))} only for infinite depth (provided water depth: {self.water_depth})."
                        )

            if self.forward_speed != 0.0:
                raise NotImplementedError(
                        f"omega={float(self.omega)} is only implemented without forward speed (provided forward speed: {self.forward_speed})."
                )


        if self.body is not None:
            if ((isinstance(self.body.mesh, CollectionOfMeshes) and len(self.body.mesh) == 0)
                    or len(self.body.mesh.faces) == 0):
                raise ValueError(f"The mesh of the body {self.body.__short_str__()} is empty.")

            panels_above_fs = self.body.mesh.faces_centers[:, 2] >= self.free_surface
            panels_below_sb = self.body.mesh.faces_centers[:, 2] <= -self.water_depth
            if (any(panels_above_fs) or any(panels_below_sb)):

                if not any(panels_below_sb):
                    issue = f"{np.count_nonzero(panels_above_fs)} panels above the free surface"
                elif not any(panels_above_fs):
                    issue = f"{np.count_nonzero(panels_below_sb)} panels below the sea bottom"
                else:
                    issue = (f"{np.count_nonzero(panels_above_fs)} panels above the free surface " +
                             f"and {np.count_nonzero(panels_below_sb)} panels below the sea bottom")

                LOG.warning(
                        f"The mesh of the body {self.body.__short_str__()} has {issue}.\n" +
                        "It has been clipped to fit inside the domain.\n" +
                        "To remove this warning, clip the mesh manually with the `immersed_part()` method."
                )

                self.body = self.body.immersed_part(free_surface=self.free_surface,
                                                    water_depth=self.water_depth)

        if self.boundary_condition is not None:
            if len(self.boundary_condition.shape) != 1:
                raise ValueError(f"Expected a 1-dimensional array as boundary_condition. Provided boundary condition's shape: {self.boundary_condition.shape}.")

            if self.boundary_condition.shape[0] != self.body.mesh.nb_faces:
                raise ValueError(
                    f"The shape of the boundary condition ({self.boundary_condition.shape})"
                    f"does not match the number of faces of the mesh ({self.body.mesh.nb_faces})."
                )

    @property
    def body_name(self):
        return self.body.name if self.body is not None else 'None'

    def _asdict(self):
        return {"body_name": self.body_name,
                "water_depth": self.water_depth,
                "omega": float(self.omega),
                "encounter_omega": float(self.encounter_omega),
                "period": float(self.period),
                "wavelength": float(self.wavelength),
                "wavenumber": float(self.wavenumber),
                "forward_speed": self.forward_speed,
                "wave_direction": self.wave_direction,
                "encounter_wave_direction": self.encounter_wave_direction,
                "rho": self.rho,
                "g": self.g}

    @staticmethod
    def _group_for_parallel_resolution(problems):
        """Given a list of problems, returns a list of groups of problems, such
        that each group should be executed in the same process to benefit from
        caching.
        """
        problems_params = pd.DataFrame([pb._asdict() for pb in problems])
        groups_of_indices = problems_params.groupby(["body_name", "water_depth", "omega", "rho", "g"]).groups.values()
        groups_of_problems = [[problems[i] for i in grp] for grp in groups_of_indices]
        return groups_of_problems

    def __str__(self):
        """Do not display default values in str(problem)."""
        parameters = [f"body={self.body.__short_str__() if self.body is not None else None}",
                      f"{self.provided_freq_type}={self.__getattribute__(self.provided_freq_type):.3f}",
                      f"water_depth={self.water_depth}"]

        if not self.forward_speed == _default_parameters['forward_speed']:
            parameters.append(f"forward_speed={self.forward_speed:.3f}")

        try:
            parameters.extend(self._str_other_attributes())
        except AttributeError:
            pass

        if not self.free_surface == _default_parameters['free_surface']:
            parameters.append(f"free_surface={self.free_surface}")
        if not self.g == _default_parameters['g']:
            parameters.append(f"g={self.g}")
        if not self.rho == _default_parameters['rho']:
            parameters.append(f"rho={self.rho}")

        return self.__class__.__name__ + "(" + ', '.join(parameters) + ")"

    def __repr__(self):
        return self.__str__()

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    def __rich_repr__(self):
        yield "body", self.body, None
        yield self.provided_freq_type, self.__getattribute__(self.provided_freq_type)
        yield "water_depth", self.water_depth, _default_parameters["water_depth"]
        try:
            yield from self._specific_rich_repr()
        except:
            pass
        yield "g", self.g, _default_parameters["g"]
        yield "rho", self.rho, _default_parameters["rho"]

    def _astuple(self):
        return (self.body, self.free_surface, self.water_depth,
                float(self.omega), float(self.period), float(self.wavenumber), float(self.wavelength),
                self.forward_speed, self.rho, self.g)

    def __eq__(self, other):
        if isinstance(other, LinearPotentialFlowProblem):
            return self._astuple() == other._astuple()
        else:
            return NotImplemented

    def __lt__(self, other):
        # Arbitrary order. Used for ordering of problems: problems with same body are grouped together.
        if isinstance(other, LinearPotentialFlowProblem):
            return self._astuple()[:9] < other._astuple()[:9]
            # Not the whole tuple, because when using inheriting classes,
            # "radiating_dof" cannot be compared with "wave_direction"
        else:
            return NotImplemented

    @property
    def depth(self):
        return self.water_depth

    @property
    def influenced_dofs(self):
        # TODO: let the user choose the influenced dofs
        return self.body.dofs if self.body is not None else set()

    def make_results_container(self):
        return LinearPotentialFlowResult(self)


class DiffractionProblem(LinearPotentialFlowProblem):
    """Particular LinearPotentialFlowProblem with boundary conditions
    computed from an incoming Airy wave."""

    def __init__(self, *,
                 body=None,
                 free_surface=_default_parameters['free_surface'],
                 water_depth=None, sea_bottom=None,
                 omega=None, period=None, wavenumber=None, wavelength=None,
                 forward_speed=_default_parameters['forward_speed'],
                 rho=_default_parameters['rho'],
                 g=_default_parameters['g'],
                 wave_direction=_default_parameters['wave_direction']):

        super().__init__(body=body, free_surface=free_surface, water_depth=water_depth, sea_bottom=sea_bottom,
                         omega=omega, period=period, wavenumber=wavenumber, wavelength=wavelength, wave_direction=wave_direction,
                         forward_speed=forward_speed, rho=rho, g=g)

        if float(self.omega) in {0.0, np.inf}:
            raise NotImplementedError(f"DiffractionProblem does not support zero or infinite frequency.")

        if self.body is not None:

            self.boundary_condition = -(
                    airy_waves_velocity(self.body.mesh.faces_centers, self)
                    * self.body.mesh.faces_normals
            ).sum(axis=1)
            # Note that even with forward speed, this is computed based on the
            # frequency and not the encounter frequency.

            if len(self.body.dofs) == 0:
                LOG.warning(f"The body {self.body.name} used in diffraction problem has no dofs!")

    def _str_other_attributes(self):
        return [f"wave_direction={self.wave_direction:.3f}"]

    def _specific_rich_repr(self):
        yield "wave_direction", self.wave_direction, _default_parameters["wave_direction"]

    def make_results_container(self, *args, **kwargs):
        return DiffractionResult(self, *args, **kwargs)


class RadiationProblem(LinearPotentialFlowProblem):
    """Particular LinearPotentialFlowProblem whose boundary conditions have
    been computed from the degree of freedom of the body."""

    def __init__(self, *, body=None,
                 free_surface=_default_parameters['free_surface'],
                 water_depth=None, sea_bottom=None,
                 omega=None, period=None, wavenumber=None, wavelength=None,
                 forward_speed=_default_parameters['forward_speed'],
                 wave_direction=_default_parameters['wave_direction'],
                 rho=_default_parameters['rho'],
                 g=_default_parameters['g'],
                 radiating_dof=None):

        self.radiating_dof = radiating_dof

        super().__init__(body=body, free_surface=free_surface, water_depth=water_depth, sea_bottom=sea_bottom,
                         omega=omega, period=period, wavenumber=wavenumber, wavelength=wavelength,
                         wave_direction=wave_direction, forward_speed=forward_speed, rho=rho, g=g)

        if self.body is not None:

            if len(self.body.dofs) == 0:
                raise ValueError(f"Body {self.body.name} does not have any degrees of freedom.")

            if self.radiating_dof is None:
                self.radiating_dof = next(iter(self.body.dofs))

            if self.radiating_dof not in self.body.dofs:
                LOG.error(f"In {self}: the radiating degree of freedom {self.radiating_dof} is not one of"
                          f"the degrees of freedom of the body.\n"
                          f"The dofs of the body are {list(self.body.dofs.keys())}")
                raise ValueError("Unrecognized degree of freedom name.")

            dof = self.body.dofs[self.radiating_dof]

            self.boundary_condition = -1j * self.encounter_omega * np.sum(dof * self.body.mesh.faces_normals, axis=1)

            if self.forward_speed != 0.0:
                if self.radiating_dof.lower() == "pitch":
                    ddofdx_dot_n = np.array([nz for (nx, ny, nz) in self.body.mesh.faces_normals])
                elif self.radiating_dof.lower() == "yaw":
                    ddofdx_dot_n = np.array([-ny for (nx, ny, nz) in self.body.mesh.faces_normals])
                elif self.radiating_dof.lower() in {"surge", "sway", "heave", "roll"}:
                    ddofdx_dot_n = 0.0
                else:
                    raise NotImplementedError(
                            "Radiation problem with forward speed is currently only implemented for a single rigid body.\n"
                            "Only radiating dofs with name in {'Surge', 'Sway', 'Heave', 'Roll', 'Pitch', 'Yaw'} are supported.\n"
                            f"Got instead `radiating_dof={self.radiating_dof}`"
                            )
                self.boundary_condition += self.forward_speed * ddofdx_dot_n


    def _astuple(self):
        return super()._astuple() + (self.radiating_dof,)

    def _asdict(self):
        d = super()._asdict()
        d["radiating_dof"] = self.radiating_dof
        return d

    def _str_other_attributes(self):
        if self.forward_speed != 0.0:
            return [f"wave_direction={self.wave_direction:.3f}, radiating_dof=\'{self.radiating_dof}\'"]
        else:
            return [f"radiating_dof=\'{self.radiating_dof}\'"]

    def _specific_rich_repr(self):
        yield "radiating_dof", self.radiating_dof

    def make_results_container(self, *args, **kwargs):
        return RadiationResult(self, *args, **kwargs)


class LinearPotentialFlowResult:

    def __init__(self, problem, forces=None, sources=None, potential=None, pressure=None):
        self.problem = problem

        self.forces = forces if forces is not None else {}
        self.sources = sources
        self.potential = potential
        self.pressure = pressure

        self.fs_elevation = {}  # Only used in legacy `get_free_surface_elevation`. To be removed?

        # Copy data from problem
        self.body               = self.problem.body
        self.free_surface       = self.problem.free_surface
        self.omega              = self.problem.omega
        self.period             = self.problem.period
        self.wavenumber         = self.problem.wavenumber
        self.wavelength         = self.problem.wavelength
        self.forward_speed      = self.problem.forward_speed
        self.wave_direction     = self.problem.wave_direction
        self.encounter_omega    = self.problem.encounter_omega
        self.encounter_period   = self.problem.encounter_period
        self.encounter_wavenumber = self.problem.encounter_wavenumber
        self.encounter_wavelength = self.problem.encounter_wavelength
        self.encounter_wave_direction = self.problem.encounter_wave_direction
        self.rho                = self.problem.rho
        self.g                  = self.problem.g
        self.boundary_condition = self.problem.boundary_condition
        self.water_depth        = self.problem.water_depth
        self.depth              = self.problem.water_depth
        self.provided_freq_type = self.problem.provided_freq_type
        self.body_name          = self.problem.body_name
        self.influenced_dofs    = self.problem.influenced_dofs

    @property
    def force(self):
        # Just an alias
        return self.forces

    __str__ = LinearPotentialFlowProblem.__str__
    __repr__ = LinearPotentialFlowProblem.__repr__
    _repr_pretty_ = LinearPotentialFlowProblem._repr_pretty_
    __rich_repr__ = LinearPotentialFlowProblem.__rich_repr__


class DiffractionResult(LinearPotentialFlowResult):

    def __init__(self, problem, *args, **kwargs):
        super().__init__(problem, *args, **kwargs)

    _str_other_attributes = DiffractionProblem._str_other_attributes
    _specific_rich_repr = DiffractionProblem._specific_rich_repr

    @property
    def records(self):
        params = self.problem._asdict()
        FK = froude_krylov_force(self.problem)
        return [dict(**params,
                     influenced_dof=dof,
                     diffraction_force=self.forces[dof],
                     Froude_Krylov_force=FK[dof])
                for dof in self.influenced_dofs]


class RadiationResult(LinearPotentialFlowResult):

    def __init__(self, problem, *args, **kwargs):
        super().__init__(problem, *args, **kwargs)
        self.radiating_dof = self.problem.radiating_dof

    _str_other_attributes = RadiationProblem._str_other_attributes
    _specific_rich_repr = RadiationProblem._specific_rich_repr

    @property
    def added_mass(self):
        return {dof: float(np.real(force)/(self.encounter_omega*self.encounter_omega)) for (dof, force) in self.forces.items()}

    @property
    def radiation_damping(self):
        return {dof: float(np.imag(force)/self.encounter_omega) for (dof, force) in self.forces.items()}

    # Aliases for backward compatibility
    added_masses = added_mass
    radiation_dampings = radiation_damping

    @property
    def records(self):
        params = self.problem._asdict()
        return [dict(params,
                     influenced_dof=dof,
                     added_mass=self.added_mass[dof],
                     radiation_damping=self.radiation_damping[dof])
                for dof in self.influenced_dofs]



"""Tools for surface integrals and hydrostatics."""
# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

from abc import ABC
import numpy as np

class SurfaceIntegralsMixin(ABC):
    # These methods need to be defined for both Mesh and CollectionOfMeshes with the exact same definitions.
    # To avoid redunduncy, they are defined here in a mixin inherited by both Mesh and CollectionOfMeshes.

    def surface_integral(self, data, **kwargs):
        """Returns integral of given data along wet surface area."""
        return np.sum(data * self.faces_areas, **kwargs)

    def waterplane_integral(self, data, **kwargs):
        """Returns integral of given data along water plane area."""
        return self.surface_integral(self.faces_normals[:,2] * data, **kwargs)

    @property
    def wet_surface_area(self):
        """Returns wet surface area."""
        return self.surface_integral(1)

    @property
    def volumes(self):
        """Returns volumes using x, y, z components of the mesh."""
        norm_coord = self.faces_normals * self.faces_centers
        return self.surface_integral(norm_coord.T, axis=1)

    @property
    def volume(self):
        """Returns volume of the mesh."""
        return np.mean(self.volumes)

    def disp_mass(self, *, rho=1000):
        return rho * self.volume

    @property
    def center_of_buoyancy(self):
        """Returns center of buoyancy of the mesh."""
        coords_sq_norm = self.faces_normals * self.faces_centers**2
        return self.surface_integral(coords_sq_norm.T, axis=1) / (2*self.volume)

    @property
    def waterplane_area(self):
        """Returns water plane area of the mesh."""
        waterplane_area = -self.waterplane_integral(1)
        return waterplane_area

    @property
    def waterplane_center(self):
        """Returns water plane center of the mesh.

        Note: Returns None if the mesh is full submerged.
        """
        waterplane_area = self.waterplane_area
        if abs(waterplane_area) < 1e-10:
            return None
        else:
            waterplane_center = -self.waterplane_integral(
                self.faces_centers.T, axis=1) / waterplane_area
            return waterplane_center[:-1]



#!/usr/bin/env python3

__all__ = ["__title__", "__description__", "__version__", "__author__", "__uri__", "__license__"]

__title__ = "capytaine"
__description__ = """Python BEM solver for linear potential flow, based on Nemoh"""

__version__ = "2.1"

__author__ = "Matthieu Ancellin"
__uri__ = "https://github.com/capytaine/capytaine"
__license__ = "GPL-3.0"


if __name__ == "__main__":
    print(__version__)



"""Computation of the impendance matrix."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

LOG = logging.getLogger(__name__)


def rao_transfer_function(dataset, dissipation=None, stiffness=None):
    """Complex-valued matrix used for the computation of the RAO.

    Parameters
    ----------
    dataset: xarray Dataset
        The hydrodynamical dataset.
        This function supposes that variables named 'inertia_matrix' and 'hydrostatic_stiffness' are in the dataset.
        Other variables can be computed by Capytaine, by those two should be manually added to the dataset.
    dissipation: array, optional
        An optional dissipation matrix (e.g. Power Take Off) to be included in the transfer function.
        Default: none.
    stiffness: array, optional
        An optional stiffness matrix (e.g. mooring stiffness) to be included in the transfer function.
        Default: none.

    Returns
    -------
    xarray DataArray
        The matrix as an array depending of omega and the degrees of freedom.
    """

    if not hasattr(dataset, 'inertia_matrix'):
        raise AttributeError('Computing the impedance matrix requires an `inertia_matrix` matrix to be defined in the hydrodynamical dataset')

    if not hasattr(dataset, 'hydrostatic_stiffness'):
        raise AttributeError('Computing the impedance matrix requires an `hydrostatic_stiffness` matrix to be defined in the hydrodynamical dataset')

    if 'encounter_omega' in dataset.coords:
        omega = dataset.coords['encounter_omega']
    else:
        omega = dataset.coords['omega']

    # ASSEMBLE MATRICES
    H = (-omega**2*(dataset['inertia_matrix'] + dataset['added_mass'])
         - 1j*omega*dataset['radiation_damping']
         + dataset['hydrostatic_stiffness'])

    if dissipation is not None:
        H = H - 1j*omega*dissipation

    if stiffness is not None:
        H = H + stiffness

    return H


def impedance(dataset, dissipation=None, stiffness=None):
    """Complex-valued mechanical impedance matrix.
    See Falnes for more theoretical details::

        @book{falnes2002ocean,
              title={Ocean Waves and Oscillating Systems: Linear Interactions Including Wave-Energy Extraction},
              author={Falnes, J.},
              isbn={9781139431934},
              url={https://books.google.com/books?id=bl1FyQjCklgC},
              year={2002},
              publisher={Cambridge University Press}
        }

    Parameters
    ----------
    dataset: xarray Dataset
        The hydrodynamical dataset.
        This function supposes that variables named 'inertia_matrix' and 'hydrostatic_stiffness' are in the dataset.
        Other variables can be computed by Capytaine, by those two should be manually added to the dataset.
    dissipation: array, optional
        An optional dissipation matrix (e.g. Power Take Off) to be included in the impedance.
        Default: none.
    stiffness: array, optional
        An optional stiffness matrix (e.g. mooring stiffness) to be included in the impedance.
        Default: none.

    Returns
    -------
    xarray DataArray
        The impedance as an array depending of omega and the degrees of freedom.
    """
    if 'encounter_omega' in dataset.coords:
        omega = dataset.coords['encounter_omega']
    else:
        omega = dataset.coords['omega']
    return 1/(-1j * omega) * rao_transfer_function(dataset, dissipation, stiffness)



import numpy as np
import xarray as xr
import capytaine as cpt

body_1 = cpt.FloatingBody(
            mesh=cpt.load_mesh("./boat_200.mar", file_format="nemoh"),
            dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0)),
        )
# If you have several rigid bodies, copy the code above to define "body_2", "body_3", etc.

all_bodies = body_1  # Replace "body_1" by "body_1 + body_2 + body_3" for multibody problem.

all_bodies = all_bodies.immersed_part()  # if your mesh has panels above the free surface, remove them

# Set up paramaters
test_matrix = xr.Dataset({
        "omega": np.linspace(0.1, 2.0, 20),  # Can also specify "period", "wavelength" or "wavenumber"
        "wave_direction": np.linspace(0, np.pi, 3),
        "radiating_dof": list(all_bodies.dofs),
        "water_depth": [np.inf],
        "rho": [1025],
        })

# Do the resolution
solver = cpt.BEMSolver()
dataset = solver.fill_dataset(test_matrix, all_bodies)

# Export to NetCDF file
from capytaine.io.xarray import separate_complex_values
separate_complex_values(dataset).to_netcdf("dataset.nc",
                                           encoding={'radiating_dof': {'dtype': 'U'},
                                                     'influenced_dof': {'dtype': 'U'}})



#!/usr/bin/env python
# coding: utf-8
"""Experimental command-line interface for Capytaine."""
# Copyright (C) 2017-2023 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import argparse

import capytaine as cpt
from capytaine.io.legacy import run_cal_file

cpt.set_logging()

parser = argparse.ArgumentParser(description="Command-line interface for Capytaine taking Nemoh.cal files as input and returning Tecplots files.")
parser.add_argument('paramfiles',
                    default=['./Nemoh.cal'],
                    nargs='*',
                    help='path of parameters files (default: ./Nemoh.cal)')


def main():
    args = parser.parse_args()
    for paramfile in args.paramfiles:
        run_cal_file(paramfile)


if __name__ == '__main__':
    main()



"""This module implements objects describing a mesh on which the free surface elevation will be computed."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from itertools import product

import numpy as np

from capytaine.meshes.meshes import Mesh

LOG = logging.getLogger(__name__)


class FreeSurface():
    """A cartesian mesh on which the free surface elevation will be computed.

    Has a :code:`mesh` attribute to behave kind of like FloatingBody when
    building of the influence matrix.

    Parameters
    ----------
    x_range: Tuple[float, float], optional
        extreme values of the mesh in the x direction
    nx: int, optional
        number of cells in the x direction
    y_range: Tuple[float, float], optional
        extreme values of the mesh in the y direction
    ny: int, optional
        number of cells in the y direction
    name: string, optional
        a name for the free surface object


    .. todo:: Generalize to non-cartesian meshes.
              In particular, it could be of interest to build meshes having the
              same symmetry as a given floating body to speed up the
              construction of the influence matrix.

    .. seealso::

        :meth:`~capytaine.bem.nemoh.Nemoh.get_free_surface_elevation`
            The main function requiring a FreeSurface object.
    """
    def __init__(self, x_range=(-50.0, 50.0), nx=10, y_range=(-50.0, 50.0), ny=10, name=None):
        self.x_range = x_range
        self.nx = nx
        self.y_range = y_range
        self.ny = ny

        if name is None:
            self.name = f"free_surface_{next(Mesh._ids)}"
        else:
            self.name = name

        self.mesh = self._generate_mesh()

    def _generate_mesh(self):
        """Generate a 2D cartesian mesh."""
        nodes = np.zeros(((self.nx+1)*(self.ny+1), 3), dtype=float)
        panels = np.zeros((self.nx*self.ny, 4), dtype=int)

        X = np.linspace(*self.x_range, self.nx+1)
        Y = np.linspace(*self.y_range, self.ny+1)
        for i, (x, y, z) in enumerate(product(X, Y, [0.0])):
            nodes[i, :] = x, y, z

        for k, (i, j) in enumerate(product(range(0, self.nx), range(0, self.ny))):
            panels[k, :] = (j+i*(self.ny+1),
                            (j+1)+i*(self.ny+1),
                            (j+1)+(i+1)*(self.ny+1),
                            j+(i+1)*(self.ny+1))

        return Mesh(nodes, panels, name=f"{self.name}_mesh")

    @property
    def area(self):
        """The total area covered by the mesh."""
        return (np.abs(self.x_range[1] - self.x_range[0])
                * np.abs(self.y_range[1] - self.y_range[0]))

    def incoming_waves(self, problem: "DiffractionProblem") -> np.ndarray:
        """Free surface elevation of the undisturbed incoming waves
        for a given diffraction problem.
        Kept for legacy, but not recommended for use.
        """
        from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
        return airy_waves_free_surface_elevation(self, problem)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

if __name__ == '__main__':

    # look for csv files in dir
    results = []
    for res_file in os.listdir('benchmarks/openmp'):
        if len(res_file) < 5 or res_file[-4:] != '.csv':
            continue
        df = pd.read_csv(os.path.join('benchmarks/openmp', res_file))
        df["filename"] = res_file
        results.append(df)

    # create dataframe
    df = pd.concat(results, axis="index")
    df.columns = df.columns.str.strip()
    df["kind"] = df["kind"].str.strip()
    df.reset_index(drop=True, inplace=True)

    df["branch"] = ''
    df["hash"] = ''
    for i in df.index:
        tmp1 = df.loc[i, "filename"][14:-4]
        tmp2 = tmp1.find('_')
        df.loc[i, "branch"] = tmp1[tmp2+1:]
        df.loc[i, "hash"] = tmp1[:tmp2]

    # plot

    mk = {'full': 'o', 'wave_only': 's', 'half_wave_only': '^'}
    hs = df['hash'].unique()
    cl = {}
    for i in range(len(hs)):
        cl[hs[i]] = 'C%1d' % (i % 10,)
    del(hs)


    plt.figure()
    for (commit, group_df) in df.groupby(["kind", "hash"]):
        plt.plot(group_df["n_threads"], group_df["elapsed_time"],
                 label='%s, %s' % (commit[1], commit[0]), ls='-',
                 marker=mk[commit[0]], c=cl[commit[1]])
    plt.xlabel("Number of threads")
    plt.ylabel("Computation time")
    plt.yscale("log")
    plt.title("OpenMP parallelization")
    plt.grid(True)
    plt.legend()

    plt.figure()
    n_threads_range = np.arange(df.n_threads.min(), df.n_threads.max()+1)
    plt.plot(n_threads_range, n_threads_range, '--', c='gray')
    for (commit, group_df) in df.groupby(["kind", "hash"]):
        ref_time = group_df.loc[group_df["n_threads"] == 1]["elapsed_time"].values[0]
        plt.plot(group_df["n_threads"], ref_time / group_df["elapsed_time"],
                 label='%s, %s' % (commit[1], commit[0]), ls='-',
                 marker=mk[commit[0]], c=cl[commit[1]])
    plt.xlabel("Number of threads")
    plt.ylabel("Speedup")
    plt.title("OpenMP parallelization (2)")
    plt.grid(True)
    plt.legend()

    plt.figure()
    for (commit, group_df) in df.groupby(["kind", "hash"]):
        ref_time = group_df.loc[group_df["n_threads"] == 1]["elapsed_time"].values[0]
        plt.plot(group_df["n_threads"], ref_time / group_df["elapsed_time"] / group_df["n_threads"],
                 label='%s, %s' % (commit[1], commit[0]), ls='-',
                 marker=mk[commit[0]], c=cl[commit[1]])
    plt.xlabel("Number of threads")
    plt.ylabel("Efficiency")
    plt.title("OpenMP parallelization (3)")
    plt.grid(True)
    plt.legend()

    plt.show()
